#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
rgmol
=====

This package provides visualization and computation of various chemical properties from DFT softwares.
Its main purpose is to compute and visualize the condensed and non-condensed linear response function from CDFT.

Documentation
-------------

The documentation can be found on `rgmol wiki <re-gr.github.io/rgmol_wiki>`_
"""


import numpy as np
import os
from rgmol.objects import *
import rgmol.extract_adf
import rgmol.extract_cube
import rgmol.extract_molden
import rgmol.extract_orca
import rgmol.extract_gaussian
import rgmol.molecular_calculations
import rgmol.calculate_CDFT
import rgmol.read_save
import rgmol.plot_pyvista
import rgmol.potential
import rgmol.grid
import rgmol.rectilinear_grid_reconstruction

__version__ = "0.1.4.1"

nprocs = 1

def set_number_procs(nprocs):
    """
    set_number_procs(nprocs)

    This functions sets the number maximum of processors that can be used by rgmol.
    Because of python shenanigans, the maximum number of processors that should be used is 8.
    Higher number of processors will slow down the calculations

    Parameters
    ----------
        nprocs : int
            the maximum number of processors

    Returns
    -------
        None
    """

    if type(nprocs) is int:
        rgmol.nprocs = nprocs
    else:
        raise TypeError('The number of processors used should be an integer. You gave a {}'.format(type(nprocs)))

set_nprocs = set_number_procs


if __name__=="__main__":
    pass





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script contains the definitions of the classes used in the whole package.
"""


import periodictable as pt
import numpy as np
from rgmol.dicts import *


class atom(object):
    """
    Atom object
    """

    def __init__(self,name_or_atomic_number,pos,properties={},**kwargs):
        """
        atom(name_or_atomic_number, pos, properties={}, **kwargs)

        Constructs atom object from its name or atomic number and positions.
        Properties can also be added in a properties dictionnary.
        If not specified, a color and a radius will be automatically attributed for the atom depending on its atomic number.

        Attributes
        ----------
            atomic_number : int
                The atomic number of the atom
            name : str
                The name of the atom
            pos : list
                The position of the atom
            nickname : str
                The nickname of the atom
            color : list
                The color of the atom used in the representations
            properties : list
                The list of the properties of the atom

        Parameters
        ----------
        name_or_atomic_number : int or str
            The atomic number or the name of the element
        pos : list
            The position of the atom
        properties : dict, optional
            The properties of the atom, if the radius is not provided, it will be taken from an already defined dictionnary.
        **kwargs :
            nickname : str
                The nickname of the atom
            color : list
                The color of the atom. If not provided, an already defined color will be used.

        Returns
        -------
            atoms : atom
                The atom object.


        """
        if type(name_or_atomic_number) is int:
            self.atomic_number = name_or_atomic_number
            self.name = pt.elements[name_or_atomic_number].symbol

        elif type(name_or_atomic_number) is str:
            self.name = name_or_atomic_number
            self.atomic_number = pt.elements.symbol(name_or_atomic_number).number

        else:
            raise TypeError("The name_or_atomic_number should be an int or a string")

        if type(pos) is list or type(pos) is np.ndarray:
            if len(pos) == 3:
                self.pos=np.array(pos)
            else: raise TypeError("The position should be a 3D list or array")
        else: raise TypeError("The position should be a 3D list or array")

        is_color = kwargs.get("color")
        if is_color:
            self.color=is_color
        else:
            self.color=(np.array(dict_color[self.name])/255).tolist()


        is_nickname = kwargs.get("nickname")
        if type(is_nickname) is str:
            self.nickname=is_nickname
        else:
            self.nickname=self.name

        self.properties = {"radius":dict_radii[self.name]}

        is_prop = kwargs.get("properties")
        if is_prop:
            if type(is_prop) is not dict:
                raise TypeError("The properties should be listed in a dict")
            for prop in is_prop:
                self.properties[prop] = is_prop[prop]

    def list_properties(self):
        """
        list_properties()

        Returns the list of the name of the properties of the atom

        Parameters
        ----------
            None

        Returns
        -------
            list_properties
                The list of the name of the properties
        """
        return list(self.properties.keys())





class molecule(object):
    """
    molecule(atoms,bonds,properties={},name=None,file="",mol_grids=None)

    Constructs a molecule from atoms.

    Parameters
    ----------
        atoms : list
            List of the atoms inside the molecule
        bonds : list
            List of the bonds
        properties : dict
            Dictionnary containing the properties of the molecules. By default : {}
        name : str, optional
            Name of the molecule
        file : str, optional
            the file from which the information are extracted
        mol_grids : mol_grids, optional
            The atomic grids of the molecule

    Returns
    -------
        molecule
            molecule object

    Attributes
    ----------
        atoms : list
        bonds : list
        properties : dict
        name : str
        file : str
        mol_grids : mol_grids
    """

    def __init__(self,atoms,bonds,properties={},name=None,file="",mol_grids=None):
        """
        molecule(atoms,bonds,properties={},name=None,file="",mol_grids=None)

        Constructs a molecule from atoms.

        Parameters
        ----------
            atoms : list
                List of the atoms inside the molecule
            bonds : list
                List of the bonds
            properties : dict
                Dictionnary containing the properties of the molecules. By default : {}
            name : str, optional
                Name of the molecule
            file : str, optional
                the file from which the information are extracted
            mol_grids : mol_grids, optional
                The atomic grids of the molecule

        Returns
        -------
            molecule
                molecule object

        Attributes
        ----------
            atoms : list
            bonds : list
            properties : dict
            name : str
            file : str
            mol_grids : mol_grids
        """

        self.name = name
        self.atoms = atoms
        self.bonds = bonds
        self.file = file
        self.mol_grids = mol_grids

        if type(properties) is not dict:
            raise TypeError("The properties should be listed in a dict")
        self.properties = properties
        Pos = []
        for atom_x in atoms:
            Pos.append(atom_x.pos)

    def list_properties(self):
        """
        list_properties()

        Returns the list of the name of the properties of the molecule

        Parameters
        ----------
            None

        Returns
        -------
            list_properties
                The list of the name of the properties
        """
        return list(self.properties.keys())


    def list_property(self,chosen_property):
        """
        list_property(chosen_property)

        Returns the list of a chosen property for all atoms.
        A list of the properties can be obtained using the list_properties() method.
        If the chosen property is "pos", it will list the position of the atoms

        Parameters
        ----------
            chosen_property : str
                The name of the chosen property. Can be "pos" for the position

        Returns
        -------
            list_property : list
                The list of the chosen property
        """
        list_property=[]
        if chosen_property=="pos":
            for atom_x in self.atoms:
                list_property.append(atom_x.pos)
            return np.array(list_property)
        for atom_x in self.atoms:
            list_property.append(atom_x.properties[chosen_property])
        return np.array(list_property)



class group_molecules(object):
    """
    group_molecules(list_molecules)

    Constructs a group of molecules objects in order to plot multiple molecules in one window

    Parameters
    ----------
        list_molecules : list
            The list containing all the molecules

    Returns
    -------
        group_molecules
            group_molecules object

    Attributes
    ----------
        list_molecules
    """

    def __init__(self,list_molecules):
        """
        group_molecules(list_molecules)

        Constructs a group of molecules objects in order to plot multiple molecules in one window

        Parameters
        ----------
            list_molecules : list
                The list containing all the molecules

        Returns
        -------
            group_molecules
                group_molecules object

        Attributes
        ----------
            list_molecules
        """
        self.molecules=list_molecules




















#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script contains the functions used to extract data from gaussian files.
THIS IS ONLY PARTIALLY IMPLEMTENTED
"""


import codecs
import numpy as np
from rgmol.objects import *
from rgmol.general_function import find_bonds


##########################
## Extraction functions ##
##########################


def _extract_fchk(file):
    """Extracts from fchk file"""

    Pos = []
    Name = []
    AO_list = []
    AO_type_list = []
    MO_list = []
    MO_energy = []
    MO_occupancy = []


    flag_name = 0
    flag_pos = 0
    flag_shell_types = 0
    flag_primitive = 0
    flag_shell_to_atom_map = 0
    flag_number_primitives = 0
    flag_contraction = 0
    flag_AO_E = 0
    flag_MO_coef = 0

    Shell_types = []
    Number_primitives = []
    Shell_to_atom_map = []
    Primitive_expo = []
    Contaction_coef = []
    MO_list_raw = []
    Contraction_coef = []

    for line in codecs.open(file, 'r',encoding="utf-8"):
        l = line.split()
        if "Atomic numbers" in line:
            flag_name = 1
        elif flag_name:
            if "Nuclear charges" in  line:
                flag_name = 0
            else:
                for atom_number in l:
                    Name.append(int(atom_number))

        elif "Number of electrons" in line:
            number_of_electrons = int(l[-1])


        elif "Current cartesian coordinates" in line:
            flag_pos = 1
            N = int(l[-1])

        elif flag_pos:
            mod = N%3
            if N>4:
                if mod == 0:
                    Pos.append(l[:3])
                    pos_mem = l[3:]
                elif mod == 1:
                    pos_mem.append(l[0])
                    Pos.append(pos_mem)
                    Pos.append(l[1:4])
                    pos_mem = [l[4]]
                else:
                    for pos in l:
                        pos_mem.append(pos)
                    Pos.append(pos_mem)
                    Pos.append(l[2:])
            else:
                if mod == 0:
                    Pos.append(l)
                elif mod == 1:
                    pos_mem.append(l[0])
                    Pos.append(pos_mem)
                    if len(l)>2:
                        Pos.append(l[1:])
                else:
                    for pos in l:
                        pos_mem.append(pos)
                    Pos.append(pos_mem)

            N-=5
            if N<1: flag_pos = 0

        elif "Shell types" in line:
            flag_shell_types = 1
        elif flag_shell_types:
            if "Number of primitives per shell" in line:
                flag_number_primitives = 1
                flag_shell_types = 0
            else:
                for shell in l:
                    Shell_types.append(shell)
        elif flag_number_primitives:
            if "Shell to atom map" in line:
                flag_shell_to_atom_map = 1
                flag_number_primitives = 0
            else:
                for num_prim in l:
                    Number_primitives.append(int(num_prim))
        elif flag_shell_to_atom_map:
            if "Primitive exponents" in line:
                flag_shell_to_atom_map = 0
                flag_primitive = 1
            else:
                for atom_map in l:
                    Shell_to_atom_map.append(int(atom_map))
        elif flag_primitive:
            if "Contraction coefficients" in line:
                flag_primitive = 0
                flag_contraction = 1
            else:
                for primitive_expo in l:
                    Primitive_expo.append(float(primitive_expo))
        elif flag_contraction:
            if "Coordinates of each shell" in line:
                flag_contraction = 0
            else:
                for contraction_coef in l:
                    Contraction_coef.append(float(contraction_coef))


        elif "Alpha Orbital Energies" in line:
            flag_AO_E = 1
        elif flag_AO_E:
            if "Alpha MO coefficients" in line:
                flag_AO_E = 0
                flag_MO_coef = 1
            else:
                for MO in l:
                    MO_energy.append(float(MO))
        elif flag_MO_coef:
            if "Orthonormal basis" in line:
                flag_MO_coef = 0
            else:
                for MO in l:
                    MO_list_raw.append(float(MO))

    mem_atom = 1
    mem_primitives = 0
    AO = []
    AO_atom = []
    AO_type = []
    convert_shell_type = {"0":"s","1":"p","-2":"d","-3":"f"}
    for atom_map,shell_type,num_primitives in zip(Shell_to_atom_map,Shell_types,Number_primitives):
        if atom_map != mem_atom:
            mem_atom = atom_map
            AO_list.append(AO_atom)
            AO_type_list.append(AO_type)
            AO_atom = []
            AO_type = []

        AO = []
        for j in range(num_primitives):
            AO.append([float(Primitive_expo[mem_primitives + j]),float(Contraction_coef[mem_primitives + j])])
        mem_primitives += num_primitives
        try:
            convert_shell_type[shell_type]
        except:
            TypeError("The type of shell {} has not been implemented".format(shell_type))

        AO_atom.append(AO)
        AO_type.append(convert_shell_type[shell_type])

    AO_list.append(AO_atom)
    AO_type_list.append(AO_type)

    s = 0



    MO_list = np.array(MO_list_raw).reshape(len(MO_energy),len(MO_energy))

    MO_occupancy = np.zeros((len(MO_energy)))
    MO_occupancy[:number_of_electrons//2] = 2

    Pos = np.array(Pos).astype("float")

    return Name, Pos, AO_list, AO_type_list, MO_list, MO_energy, MO_occupancy




def extract_transition(file,mol=None):
    """Extracts from log"""

    conversion_ev_to_ha = 1/27.2114

    flag_excited_states = 0
    flag_doing_excited_states = 0
    flag_pause = 0

    transition_energy = []
    transition_factor_list = []
    transition_list = []

    for line in codecs.open(file, 'r',encoding="utf-8"):
        l = line.split()
        if "Excitation energies and oscillator strengths" in line:
            flag_excited_states = 1
            flag_pause = 1
        elif flag_pause:
            flag_pause = 0

        elif flag_excited_states:
            if "Excited State" in line:
                flag_doing_excited_states = 1
                transition_energy.append(float(l[4])*conversion_ev_to_ha)
                transition_MO = []
                transition_MO_coef = []
            elif flag_doing_excited_states:
                if "SavETr" in line:
                    flag_excited_states = 0
                    transition_factor_list.append(transition_MO_coef)
                    transition_list.append(transition_MO)
                elif "This state" in line or len(line)<3:
                    flag_doing_excited_states = 0
                    transition_factor_list.append(transition_MO_coef)
                    transition_list.append(transition_MO)
                else:
                    transition_MO.append([int(l[0])-1,int(l[2])-1])
                    transition_MO_coef.append([float(l[-1])])

    array_sort = np.argsort(transition_energy)
    transition_energy_sorted = []
    transition_list_sorted = []
    transition_factor_list_sorted = []



    for sorting in array_sort:
        transition_energy_sorted.append(transition_energy[sorting])
        transition_list_sorted.append(transition_list[sorting])
        transition_factor_list_sorted.append(transition_factor_list[sorting])

    if mol:
        mol.properties["transition_energy"] = transition_energy_sorted
        mol.properties["transition_list"] = transition_list_sorted
        mol.properties["transition_factor_list"] = transition_factor_list_sorted

    return  transition_energy_sorted, transition_list_sorted, transition_factor_list_sorted



def extract_fchk(file,do_find_bonds=0):
    """extracts from fchk and creates mol"""
    if file[-4:] == "fchk":
        atom_names, atom_position, AO_list, AO_type_list, MO_list, MO_energy, MO_occupancy = _extract_fchk(file)
    list_atoms = []
    nicknaming = 1

    for prop in zip(atom_names,atom_position):
        atom_x = atom(prop[0],prop[1],nickname=str(nicknaming))
        list_atoms.append(atom_x)
        nicknaming += 1

    mol = molecule(list_atoms,[],file=file,properties={"AO_list":AO_list,"AO_type_list":AO_type_list,"MO_list":MO_list,"MO_energy":MO_energy,"MO_occupancy":MO_occupancy})

    if do_find_bonds:
        mol.bonds = find_bonds(mol)

    return mol





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script adds functions, and methods to compute potential from a molecule.
"""

import os
import numpy as np
import scipy as sp
from rgmol.objects import *
# from rgmol.molecular_calculations import create_coordinates_from_voxel
# from rgmol.molecular_calculations import create_voxel_from_molecule

def set_external_potential(self,ext_mol,grid_points=(100,100,100),delta=5,recalculate=False):
    """

    """

    if not recalculate:
        #tries to find the cube file of the
        pass

    print("Computing the external potential. This may take a while")

    if not "voxel_origin" in self.properties:
        voxel_origin,voxel_matrix = create_voxel_from_molecule(self,grid_points,delta=delta)
        self.properties["voxel_origin"] = voxel_origin
        self.properties["voxel_matrix"] = voxel_matrix
        self.properties["grid_points"] = grid_points

    voxel_origin = self.properties["voxel_origin"]
    voxel_matrix = self.properties["voxel_matrix"]
    dV = voxel_matrix[0][0] * voxel_matrix[1][1] * voxel_matrix[2][2]


    r = create_coordinates_from_voxel(grid_points,voxel_origin,voxel_matrix)

    total_external_potential = np.zeros(grid_points)

    for ext_atom in ext_mol.atoms:
        pos_pot = np.reshape(ext_atom.pos,(3,1,1,1))
        charge = ext_atom.atomic_number

        dist = np.sum((r-pos_pot)**2,axis=0)**(1/2)

        total_external_potential += charge / dist

    ext_electron_density = ext_mol.calculate_electron_density(grid_points=grid_points,delta=delta)

    ext_voxel_origin = ext_mol.properties["voxel_origin"]
    ext_voxel_matrix = ext_mol.properties["voxel_matrix"]

    rp = create_coordinates_from_voxel(grid_points,ext_voxel_origin,ext_voxel_matrix)

    for x in range(grid_points[0]):
        for y in range(grid_points[1]):
            for z in range(grid_points[2]):
                dist = np.sum((r[:,x,y,z].reshape((3,1,1,1)) - rp)**2,axis=0)**(1/2)
                total_external_potential[x,y,z] = total_external_potential[x,y,z] + np.sum(ext_electron_density/dist)

    print("Finished computing the external potential")


  #
  # 		for file in D[k*mult:(k+1)*mult]:
  # 			a=th.Thread(target=deconvo,args=(peaks,file[2:-4],chem+file,dim,perfect,name_sample+"_"+file[2:-4],A[1],height))
  # 			a.start()
  # 		a.join()
  #
  #






molecule.set_external_potential = set_external_potential


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script adds writing and reading functions
"""

import os
import time
import numpy as np
import scipy as sp
import zipfile as zf
from rgmol.objects import *
import rgmol
import rgmol.grid


def _get_file_location(file):
    """gets the location of the file of the output of the molecule"""
    file_splitted = file.split("/")[:-1]
    file_location = ""

    for f in file_splitted:
        file_location += f
        file_location +="/"

    return file_location

def _write_vector_for_cube(Lines,vector):
    """Writes a vector in a file"""
    for elem in vector:
        long = len(str(abs(int(elem))))
        Lines.append(" "*(4-long))
        Lines.append("{:9.6f}".format(elem))

def _write_data_for_cube(Lines,vector):
    """Writes a vector in a file"""
    line = ""
    for elem in vector:
        m = int(np.sign(elem)/2-1/2)
        line += " "*(3+m)
        line += "{:9.5e}".format(elem)
    Lines.append(line)

def _write_voxel(file_name,grid_points,delta):
    """Writes the voxel parameters in a file"""

    with open(file_name,"w") as file:
        file.write("Grid Points :\n")
        file.write("{} {} {}\n".format(grid_points[0],grid_points[1],grid_points[2]))
        file.write("Delta :\n")
        file.write("{}".format(delta))


def _read_voxel(file_name):
    """Reads the voxel parameters from a file written using the write_voxel function"""

    count = 0
    voxel_matrix=[]
    for line in open(file_name,"r"):
        lsplit = line.split()
        if count == 1:
            grid_points = [int(lsplit[0]),int(lsplit[1]),int(lsplit[2])]
        elif count == 3:
            delta = float(lsplit[0])
        count+=1
    return grid_points,delta


def write_cube(self,cube,file_name,description="cube"):
    """
    write_cube(cube,file_name,description="cube")

    Writes a cube file of a 3D vector inside a folder called rgmol which will be located inside the folder containing the molecule.

    Parameters
    ----------
        cube : ndarray
            The 3D vector to be stored in the cube file
        file_name : str
            The name of the file
        description : str
            Description of the file inside the cube file

    Returns
    -------
        None
            The cube file is written
    """
    Lines = []
    Lines.append("Cube data generated by rgmol\n")
    Lines.append(description+"\n")

    #Number atoms and voxel origin
    num_atoms = str(len(self.atoms))
    voxel_origin = self.properties["voxel_origin"]

    Lines.append(" "*(5-len(num_atoms)) +num_atoms)
    _write_vector_for_cube(Lines,voxel_origin)

    Lines.append("\n")

    #Number of points and voxel matrix
    voxel_matrix = self.properties["voxel_matrix"]
    grid_points = self.properties["grid_points"]

    grid_points_0 = str(grid_points[0])
    Lines.append(" "*(5-len(grid_points_0)) + grid_points_0)
    _write_vector_for_cube(Lines,voxel_matrix[0])
    Lines.append("\n")
    grid_points_1 = str(grid_points[1])
    Lines.append(" "*(5-len(grid_points_1)) + grid_points_1)
    _write_vector_for_cube(Lines,voxel_matrix[1])
    Lines.append("\n")
    grid_points_2 = str(grid_points[2])
    Lines.append(" "*(5-len(grid_points_2)) + grid_points_2)
    _write_vector_for_cube(Lines,voxel_matrix[2])
    Lines.append("\n")

    #Write atom positions
    for atom in self.atoms:
        atomic_number = atom.atomic_number
        Lines.append(" "*(5-len(str(atomic_number)))+str(atomic_number))
        Lines.append(" "*(4-len(str(atomic_number)))+"{:9.6f}".format(atomic_number))
        _write_vector_for_cube(Lines,atom.pos)
        Lines.append("\n")

    #Write cube
    for vec_x in cube:
        for vec_y in vec_x:
            line_length = grid_points[2]
            count = 0
            while count < line_length:
                _write_data_for_cube(Lines,vec_y[count:count+6])
                Lines.append('\n')
                count+=6

    with open(file_name,"w") as file:
        file.writelines(Lines)

# def write_txt(self,cube,file_name):
#     """
#     write_txt(cube,file_name)
#
#     This functions writes a txt file containing a 3D array.
#     This method is not the preferred one, just use mol.save()
#
#     Parameters
#     ----------
#         cube : ndarray
#             The cube file to save
#         file_name : str
#             The name of the file
#
#     Returns
#     -------
#         None
#             The file is written
#     """
#
#     grid_points = self.properties["grid_points"]
#     nx,ny,nz = grid_points
#     comments = ""
#     for coord in grid_points:
#         comments += str(coord) + " "
#
#     np.savetxt(file_name,cube.reshape((nx*ny,nz)),header=comments,fmt="%3.6e")


molecule.write_cube = write_cube
# molecule.write_txt = write_txt



#########################################
## Adding save functions for molecules ##
#########################################


def _save_kernel(mol,file_location,rgmol_folder,kernel_name,eigenvectors):
    """
    This functions saves the eigenmodes of a kernel
    """

    try: os.mkdir(file_location + rgmol_folder+"//"+kernel_name)
    except: pass
    for eigenvector in range(len(eigenvectors)):
        file_name = "eigenvector_{}.npy".format(eigenvector)
        np.save(file_location+rgmol_folder+"//"+kernel_name+"//"+file_name,eigenvectors[eigenvector])



def save(self,kernel,append_name=""):
    """
    save(kernel,append_name="")

    This function saves the eignemodes of the already computed kernels in a folder called rgmol.
    The rgmol folder is located in the same folder as the input file used for the molecule.

    Parameters
    ----------
        append_name : str, optional
            A string to append to the name of the rgmol folder.

    Returns
    -------
        None
            The files are written in the rgmol folder

    """

    file_location = _get_file_location(self.file)
    rgmol_folder = "rgmol" + append_name

    try: os.mkdir(file_location + rgmol_folder)
    except: pass
    try: os.mkdir(file_location + "temp")
    except: pass

    print("############################")
    print("# Writing the eigenvectors #")
    print("############################")
    time_before_writing = time.time()
    if kernel == "linear_response_function":
        if "Reconstructed_linear_response_eigenvectors" in self.properties:
            linear_response_eigenvalues = self.properties["linear_response_eigenvalues"]
            contribution_linear_response_eigenvectors = self.properties["contribution_linear_response_eigenvectors"]
            Reconstructed_linear_response_eigenvectors = self.properties["Reconstructed_linear_response_eigenvectors"]

            np.savetxt(file_location + rgmol_folder + "//linear_response_eigenvalues.txt",linear_response_eigenvalues,comments="#",header="Linear Response Eigenvalues")
            np.savetxt(file_location + rgmol_folder + "//contribution_linear_response.txt",contribution_linear_response_eigenvectors,comments="#",header="Contribution of transition densities on eigenmodes")
            _save_kernel(self,file_location,rgmol_folder,"linear_response_function",Reconstructed_linear_response_eigenvectors)
            grid_points = self.properties["grid_points"]
            delta = self.properties["delta"]
            _write_voxel(file_location + rgmol_folder +"//voxel_parameters.txt",grid_points,delta)


        if "linear_response_eigenvectors" in self.properties:
            linear_response_eigenvectors = self.properties["linear_response_eigenvectors"]
            linear_response_eigenvalues = self.properties["linear_response_eigenvalues"]
            contribution_linear_response_eigenvectors = self.properties["contribution_linear_response_eigenvectors"]
            np.savetxt(file_location + rgmol_folder + "//linear_response_eigenvalues.txt",linear_response_eigenvalues,comments="#",header="Linear Response Eigenvalues")
            np.savetxt(file_location + rgmol_folder + "//contribution_linear_response.txt",contribution_linear_response_eigenvectors,comments="#",header="Contribution of transition densities on eigenmodes")
            np.save(file_location + rgmol_folder + "//linear_response_eigenvectors",linear_response_eigenvectors)

    elif kernel == "softness_kernel":
        if "Reconstructed_softness_kernel_eigenvectors" in self.properties:
            softness_kernel_eigenvalues = self.properties["softness_kernel_eigenvalues"]
            contribution_softness_kernel_eigenvectors = self.properties["contribution_softness_kernel_eigenvectors"]
            Reconstructed_softness_kernel_eigenvectors = self.properties["Reconstructed_softness_kernel_eigenvectors"]

            np.savetxt(file_location + rgmol_folder + "//softness_kernel_eigenvalues.txt",softness_kernel_eigenvalues,comments="#",header="Linear Response Eigenvalues")
            np.savetxt(file_location + rgmol_folder + "//contribution_softness_kernel.txt",contribution_softness_kernel_eigenvectors,comments="#",header="Contribution of transition densities on eigenmodes")
            _save_kernel(self,file_location,rgmol_folder,"softness_kernel",Reconstructed_softness_kernel_eigenvectors)
            grid_points = self.properties["grid_points"]
            delta = self.properties["delta"]
            _write_voxel(file_location + rgmol_folder +"//voxel_parameters.txt",grid_points,delta)


        if "softness_kernel_eigenvectors" in self.properties:
            softness_kernel_eigenvectors = self.properties["softness_kernel_eigenvectors"]
            softness_kernel_eigenvalues = self.properties["softness_kernel_eigenvalues"]
            contribution_softness_kernel_eigenvectors = self.properties["contribution_softness_kernel_eigenvectors"]
            np.savetxt(file_location + rgmol_folder + "//softness_kernel_eigenvalues.txt",softness_kernel_eigenvalues,comments="#",header="Linear Response Eigenvalues")
            np.savetxt(file_location + rgmol_folder + "//contribution_softness_kernel.txt",contribution_softness_kernel_eigenvectors,comments="#",header="Contribution of transition densities on eigenmodes")
            np.save(file_location + rgmol_folder + "//softness_kernel_eigenvectors",softness_kernel_eigenvectors)

    else:
        raise ValueError("This kernel {} has not been implemented. Please choose linear_response_function or softness kernel".format(kernel))

    os.rmdir(file_location + "temp")

    print("#####################################")
    print("# Finished Writing the eigenvectors #")
    print("# in {:3.3f} #".format(time.time()-time_before_writing))
    print("#####################################")



def read(self,append_name="",nb_eigen=0):
    """
    read()

    This function reads the files located in the rgmol folder where the molecule is saved.

    Parameters
    ----------
        append_name : str, optional
            A string to append to the name of the rgmol folder.
        nb_eigen : int, optional
            How many eigenmodes will be extracted

    Returns
    -------
        None
            The properties will be put inside molecule.properties

    """

    file_location = _get_file_location(self.file)
    rgmol_folder = "rgmol" + append_name

    try: os.mkdir(file_location + "temp")
    except: pass
    listdir = os.listdir(file_location)
    if not rgmol_folder in listdir:
        raise TypeError("No {} folder were found in the location where the input file of the molecule is.".format(rgmol_folder))

    listdir_rgmol = os.listdir(file_location + rgmol_folder)

    if "voxel_parameters.txt" in listdir_rgmol:
        grid_points,delta = _read_voxel(file_location+rgmol_folder+"//voxel_parameters.txt")
        self.properties["grid_points"] = grid_points
        self.properties["delta"] = delta
        r,c = rgmol.grid.create_cubic_grid_from_molecule(self,grid_points=grid_points,delta=delta)
        self.properties["Reconstructed_coords"] = c

    else:
        print("No representation grid parameters were found, the extraction will proceed anyway")


    if "linear_response_function" in listdir_rgmol:
        print("#############################")
        print("# Extracting the eigenmodes #")
        print("#############################")
        time_before_extract = time.time()

        linear_response_eigenvectors = []

        listdir = os.listdir(file_location+rgmol_folder+"//linear_response_function")
        listdir_int = [int(files.split(".")[0][12:]) for files in listdir]

        arr_sort = np.argsort(listdir_int)
        listdir_sorted = []
        for index in arr_sort:
            listdir_sorted.append(listdir[index])

        if nb_eigen:
            listdir_sorted = listdir_sorted[:nb_eigen]

        for file in listdir_sorted:
            linear_response_eigenvectors.append(np.load(file_location+rgmol_folder+"//linear_response_function//"+file))
        self.properties["Reconstructed_linear_response_eigenvectors"] = linear_response_eigenvectors

        print("######################################")
        print("# Finished Extracting the eigenmodes #")
        print("# in {:3.3f} #".format(time.time()-time_before_extract))
        print("######################################")

    if "linear_response_eigenvectors.npy" in listdir_rgmol:
        print("############################################")
        print("# Extracting the eigenmodes on atomic grid #")
        print("############################################")
        time_before_extract = time.time()

        linear_response_eigenvectors = np.load(file_location+rgmol_folder+"//linear_response_eigenvectors.npy")
        self.properties["linear_response_eigenvectors"] = linear_response_eigenvectors

        print("#####################################################")
        print("# Finished Extracting the eigenmodes on atomic grid #")
        print("# in {:3.3f} #".format(time.time()-time_before_extract))
        print("#####################################################")

        #Recreate grid
        rgmol.grid.create_grid_from_mol(self,N_r=None,d_leb=None,zeta_list=None,alpha_list=None)



    if "softness_kernel" in listdir_rgmol:
        print("#############################")
        print("# Extracting the eigenmodes #")
        print("#############################")
        time_before_extract = time.time()

        softness_kernel_eigenvectors = []

        listdir = os.listdir(file_location+rgmol_folder+"//softness_kernel")
        listdir_int = [int(files.split(".")[0][12:]) for files in listdir]

        arr_sort = np.argsort(listdir_int)
        listdir_sorted = []
        for index in arr_sort:
            listdir_sorted.append(listdir[index])

        for file in listdir_sorted:
            softness_kernel_eigenvectors.append(np.load(file_location+rgmol_folder+"//softness_kernel//"+file))
        self.properties["Reconstructed_softness_kernel_eigenvectors"] = softness_kernel_eigenvectors

        print("######################################")
        print("# Finished Extracting the eigenmodes #")
        print("# in {:3.3f} #".format(time.time()-time_before_extract))
        print("######################################")

    if "softness_kernel_eigenvectors.npy" in listdir_rgmol:
        print("############################################")
        print("# Extracting the eigenmodes on atomic grid #")
        print("############################################")
        time_before_extract = time.time()

        softness_kernel_eigenvectors = np.load(file_location+rgmol_folder+"//softness_kernel_eigenvectors.npy")
        self.properties["softness_kernel_eigenvectors"] = softness_kernel_eigenvectors

        print("#####################################################")
        print("# Finished Extracting the eigenmodes on atomic grid #")
        print("# in {:3.3f} #".format(time.time()-time_before_extract))
        print("#####################################################")

        #Recreate grid
        rgmol.grid.create_grid_from_mol(self,N_r=None,d_leb=None,zeta_list=None,alpha_list=None)


    if "linear_response_eigenvalues.txt" in listdir_rgmol:
        linear_response_eigenvalues = np.loadtxt(file_location + rgmol_folder + "//linear_response_eigenvalues.txt")
        contribution_linear_response = np.real(np.loadtxt(file_location + rgmol_folder + "//contribution_linear_response.txt",dtype=np.complex_))

        self.properties["linear_response_eigenvalues"] = linear_response_eigenvalues
        self.properties["contribution_linear_response_eigenvectors"] = contribution_linear_response

    if "softness_kernel_eigenvalues.txt" in listdir_rgmol:
        softness_kernel_eigenvalues = np.loadtxt(file_location + rgmol_folder + "//softness_kernel_eigenvalues.txt")
        contribution_softness_kernel = np.loadtxt(file_location + rgmol_folder + "//contribution_softness_kernel.txt")

        self.properties["softness_kernel_eigenvalues"] = softness_kernel_eigenvalues
        self.properties["contribution_softness_kernel_eigenvectors"] = contribution_softness_kernel


molecule.save = save
molecule.read = read


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script adds functions, and methods to the molecule objects.
These methods allow the plotting of chemical properties using pyvista.
"""

import numpy as np
import pyvista
import matplotlib.pyplot as plt
from rgmol.objects import *
import rgmol.molecular_calculations
import rgmol.grid
import rgmol.rectilinear_grid_reconstruction
from ._functions import *

###############################
## Defines class for sliders ##
###############################

class _slider:
    def __init__(self,func,value,cutoff):
        self.func = func
        self.value = value
        self.cutoff = cutoff

    def __call__(self, slider, value):
        if slider == "cutoff":
            self.cutoff = value
        else:
            self.value = value
        self.func(self.value,self.cutoff)

class _three_buttons:
    def __init__(self,func,value,mode_D,mode_P,mode_M):
        self.func = func
        self.value = value
        self.mode_D = mode_D
        self.mode_P = mode_P
        self.mode_M = mode_M

    def __call__(self,mode,value):
        if mode=="value":
            self.value = value
        elif mode=="d":
            self.mode_D = value
        elif mode=="p":
            self.mode_P = value
        elif mode=="m":
            self.mode_M = value
        self.func(self.value,self.mode_D,self.mode_P,self.mode_M)

##############################################
## Adding Plotting Methods for Atomic Grids ##
##############################################


def plot_atomic_grid(self,N_r_list=None,d_leb_list=None,zeta_list=None,alpha_list=None,opacity=0.5):
    """
    plot_atomic_grid(N_r_list=None,d_leb_list=None,zeta_list=None,alpha_list=None,opacity=0.5)

    Do a plot of the grid of the molecule, each point of the grid is represented as a point.
    The color of each point is the same as the atom that is at the center of the atomic grid
    For more information on the grids, please check :doc:`here<../grid/create_atomic_grid>`

    Parameters
    ----------
        N_r_list : list
            The number of radial points that will be taken for each atom repectively
        d_leb_list : list
            The degree of the Lebedev quadrature that will be taken for each atom respectively
        zeta_list : list
            The zeta parameter that will be taken for each atom repectively
        alpha_list : list
            The alpha parameter that will be taken for each atom repectively

    Returns
    -------
        None
            The plotter will show
    """
    if not self.mol_grids:
        rgmol.grid.create_grid_from_mol(self)


    plotter = pyvista.Plotter()


    for grid,atom in zip(self.mol_grids.grids,self.atoms):
        coords = grid.xyz_coords
        num_points = grid.number_points
        coords = coords.reshape((3,num_points)).transpose()
        points = pyvista.PolyData(coords)
        plotter.add_points(points,render_points_as_spheres=True,point_size=40,opacity=opacity,color=atom.color)

    for atom_x in self.atoms:
        atom_x.plot(plotter,opacity=1.0,factor=1)



    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.show(full_screen=False)
    return


def plot_on_atomic_grid(self,arr,N_r_list=None,d_leb_list=None,zeta_list=None,alpha_list=None):
    """
    plot_on_atomic_grid(self,arr,N_r_list=None,d_leb_list=None,zeta_list=None,alpha_list=None)

    Do a plot of an array on the atomic grids.
    The array should be of the shape : (N_grids, N_r, N_ang)
    For more information on the grids, please check :doc:`here<../grid/create_atomic_grid>`

    Parameters
    ----------
        arr : ndarray
            The array to be plotted, the shape must be (N_grids, N_r, N_ang)
        N_r_list : list, optional
            The number of radial points that will be taken for each atom repectively
        d_leb_list : list, optional
            The degree of the Lebedev quadrature that will be taken for each atom respectively
        zeta_list : list, optional
            The zeta parameter that will be taken for each atom repectively
        alpha_list : list, optional
            The alpha parameter that will be taken for each atom repectively

    Returns
    -------
        None
            The plotter will show
    """
    if not self.mol_grids:
        rgmol.grid.create_grid_from_mol(self,N_r_list=N_r_list,d_leb_list=d_leb_list,zeta_list=zeta_list,alpha_list=alpha_list)


    plotter = pyvista.Plotter()


    maxx = 0
    for grid,array in zip(self.mol_grids.grids,arr):
        wn = grid.wn
        dV = grid.dV
        val = np.max(abs(array*wn*dV))
        if val > maxx:
            maxx = val

    for grid,atom,array_to_plot in zip(self.mol_grids.grids,self.atoms,arr):
        coords = grid.xyz_coords
        num_points = grid.number_points
        coords = coords.reshape((3,num_points)).transpose()
        points = pyvista.PolyData(coords)
        wn = grid.wn
        dV = grid.dV

        points = plotter.add_points(points,render_points_as_spheres=True,point_size=40,opacity=[1.0,0.4,0.0,0.4,1.0],scalars=array_to_plot*wn*dV,clim=(-maxx,maxx),cmap="rainbow4")


    for atom_x in self.atoms:
        atom_x.plot(plotter,opacity=0.3,factor=0.5)
    bonds_plotting(plotter,self.bonds,self.list_property("pos"),self.list_property("radius"),factor=.5)



    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.show(full_screen=False)
    return

molecule.plot_atomic_grid = plot_atomic_grid
molecule.plot_on_atomic_grid = plot_on_atomic_grid


########################################
## Adding Plotting Methods for Atoms  ##
########################################



def plot(self,plotter,plotted_property="radius",opacity=1,factor=1):
    """
    plot(plotter,plotted_property="radius",opacity=1,factor=1)

    Plot a property of the atom on the plotter using pyvista

    Parameters :
    ------------
        plotter : pyvista.plotter
            The plotter object from pyvita on which the atom will be plotted. It can be easily defined using plotter = pyvista.Plotter()
        plotted_property : string, optional
            The property to be plotted. By default the radius is plotted.
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1

    Returns :
    ---------
        None
            The atom is plotted on the plotter object
    """
    plot_atom(plotter,self,plotted_property=plotted_property,opacity=opacity,factor=factor)



def plot_vector(self,plotter,vector,opacity=1,factor=1):
    """
    plot_vector(plotter,plotted_property="radius",opacity=1,factor=1)

    Plot a value of a vector on the position of the atom using pyvista

    Parameters :
    ------------
        plotter : pyvista.plotter
            The plotter object from pyvita on which the atom will be plotted. It can be easily defined using plotter = pyvista.Plotter()
        vector : float
            The value to be plotted
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1

    Returns :
    ---------
        None
            The atom is plotted on the plotter object
    """
    plot_vector_atom(plotter,self,vector,opacity=opacity,factor=factor)


atom.plot = plot
atom.plot_vector = plot_vector


############################################
## Adding Plotting Methods for Molecules  ##
############################################



def plot(self,plotter,plotted_property="radius",opacity=1,factor=1,show_bonds=True):
    """
    plot(plotter,plotted_property="radius",opacity=1,factor=1,show_bonds=True)

    Plot a property for the entire molecule

    Parameters
    ----------
        plotter : pyvista.plotter
            The plotter object from pyvita on which the atom will be plotted. It can be easily defined using plotter = pyvista.Plotter()
        plotted_property : float, optional
            The property to be plotted. By default the radius
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        show_bonds : bool, optional
            Chose to show the bonds between the atoms or not. By default True

    Returns
    -------
        None
            The atom is plotted on the plotter object
    """
    for atom_x in self.atoms:
        atom_x.plot(plotter,plotted_property=plotted_property,opacity=opacity,factor=factor)
    if show_bonds:
        bonds_plotting(plotter,self.bonds,self.list_property("pos"),self.list_property(plotted_property),factor=factor)
    return



def plot_vector(self,plotter,vector,opacity=1,factor=1):
    """
    plot_vector(plotter,vector,opacity=1,factor=1,show_bonds=True)

    Plot a vector on each atom for the entire molecule

    Parameters
    ----------
        plotter : pyvista.plotter
            The plotter object from pyvita on which the atom will be plotted. It can be easily defined using plotter = pyvista.Plotter()
        vector : list or ndarray
            The vector to be plotted. The length should be equal to the number of atoms.
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        show_bonds : bool, optional
            Choose to show the bonds between the atoms or not. By default True

    Returns
    -------
        None
            The atom is plotted on the plotter object
    """
    for atom_x in range(len(self.atoms)):
        self.atoms[atom_x].plot_vector(plotter,vector[atom_x],opacity=opacity,factor=factor)
    return


def plot_radius(self,opacity=1,factor=.4,show_bonds=True,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_radius(opacity=1,factor=.4,show_bonds=True,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot the radius of the entire molecule

    Parameters
    ----------
    opacity : float, optional
        The opacity of the plot. By default equals to 1
    factor : float, optional
        The factor by which the plotted_property will be multiplied. By default equals to 1
    show_bonds : bool, optional
        Choose to show the bonds between the atoms or not. By default True
    screenshot_button : bool, optional
        Adds a screenshot button. True by default
    window_size_screenshot : tuple, optional
        The size of the screenshots. By default (1000,1000)

    Returns
    -------
    None
        The plotter should display when using this function
    """
    plotter = pyvista.Plotter()
    for atom_x in self.atoms:
        atom_x.plot(plotter,opacity=opacity,factor=factor)
    if show_bonds:
        bonds_plotting(plotter,self.bonds,self.list_property("pos"),self.list_property("radius"),factor=factor)
    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)
    return


def plot_property(self,plotted_property,opacity=.4,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_property(plotted_property,opacity=.4,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot a property for the entire molecule

    Parameters
    ----------
        plotted_property : float, optional
            The property to be plotted. By default the radius
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """
    X = self.properties[plotted_property]
    plotter = pyvista.Plotter()

    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)
    self.plot_vector(plotter,X,opacity=opacity,factor=factor)
    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)
    return


def plot_condensed_kernel(self,kernel,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_condensed_kernel(kernel,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plots a condensed kernel for a molecule. One can navigate through the vectors using a slider.

    Parameters
    ----------
        kernel : str
            The kernel to be diagonalized and plotted
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """
    X = self.properties[kernel]
    ncols = len(X)

    plotter = pyvista.Plotter()
    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)
    def create_mesh_diagonalized_kernel(value):
        vector_number = int(round(value))
        plot_single_atom(plotter,self.atoms[vector_number-1],factor=factor_radius*1.5)
        self.plot_vector(plotter,X[:,vector_number-1],opacity=opacity,factor=factor)


    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(create_mesh_diagonalized_kernel, [1, len(X)],value=1,title="Vector", fmt="%1.0f")
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)
    return

def plot_diagonalized_condensed_kernel(self,kernel,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_diagonalized_condensed_kernel(kernel,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,screenshot_button=True,window_size_screenshot=(1000,1000))

    Diagonalize and plot a condensed kernel for a molecule. One can navigate through the eigenmodes using a slider.

    Parameters
    ----------
        kernel : str
            The kernel to be diagonalized and plotted
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """
    X = self.properties[kernel]
    Xvp,XV = np.linalg.eigh(X)
    ncols = len(X)

    plotter = pyvista.Plotter()
    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)
    def create_mesh_diagonalized_kernel(value):
        vector_number = int(round(value))
        self.plot_vector(plotter,XV[:,vector_number-1],opacity=opacity,factor=factor)
        plotter.add_text(text=r"eigenvalue = "+'{:3.3f} (a.u.)'.format(Xvp[vector_number-1]),name="eigenvalue")


    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(create_mesh_diagonalized_kernel, [1, len(XV)],value=1,title="Eigenvector", fmt="%1.0f")
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)
    return

#########################
## General 3D plotting ##
#########################

def plot_isodensity(self,plotted_isodensity="cube",opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_isodensity(plotted_isodensity="cube",opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot an isodensity

    Parameters
    ----------
        plotted_isodensity : str, optional
            The isodensity to be plotted. By default "cube"
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """
    plotter = pyvista.Plotter()
    r,coords = rgmol.grid.create_cubic_grid_from_molecule(self,grid_points=np.shape(self.properties[plotted_isodensity]))
    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius,show_bonds=True)
    def create_mesh_cube(value):
        plot_cube(plotter,coords,self.properties[plotted_isodensity],opacity=opacity,factor=factor,cutoff=value)

    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(create_mesh_cube, [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)

    return




def plot_multiple_isodensities(base_name_file,list_files,plotted_isodensity="cube",delimiter="&",opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_multiple_isodensities(range_files,plotted_isodensity="cube",delimiter=" ",opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot multiple isodensities, each one can be selected using a slider. The delimiter is replaced in the name file by number in the range_file to load multiple files
    Only one delimiter should be used in the base_name_file

    Parameters
    ----------
        base_name_file : str
            The base name of the files. If one wants to view files called H2CO.mo0a.cube, H2CO.mo1a.cube ...
            One should set the base_name_file as H2CO.mo&a.cube
        list_files : tuple
            The list of strings to replace the delimiter with
        plotted_isodensity : str, optional
            The isodensity to be plotted. By default "cube"
        delimiter : str, optional
            The
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    plotter = pyvista.Plotter()

    def open_isodensity(value,cutoff):
        value = int(round(value))
        splitted_name = base_name_file.split(delimiter)
        file = splitted_name[0] + str(value) + splitted_name[1]
        mol = rgmol.extract_cube.extract(file,do_find_bonds=1)

        if with_radius:
            mol.plot(plotter,factor=factor_radius,opacity=opacity_radius,show_bonds=True)
        plot_cube(plotter,mol.properties["voxel_origin"],mol.properties["voxel_matrix"],mol.properties["cube"],opacity=opacity,factor=factor,cutoff=cutoff)

    slider_function = _slider(open_isodensity,1,cutoff)

    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(lambda value: slider_function("AO",value), [int(list_files[0]),len(list_files)],value=1,title="Number", fmt="%1.0f")
    plotter.add_slider_widget(lambda value: slider_function("cutoff",value), [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)
    return




############################################
## Plotting Atomic / Molecular Properties ##
############################################

def plot_AO(self,grid_points=(80,80,80),delta=5,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_AO(grid_points=(80,80,80),delta=5,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot the Atomic Orbitals of a molecule.
    Because no calculations are done, the Atomic Orbitals will be calculated on the representative grid.
    The representative grid is a cubic grid around the molecule.
    The delta defines the length to be added to the extremities of the position of the atoms.
    The order of the Atomic Orbitals is defined in the molden file

    Parameters
    ----------
        grid_points : list of 3, optional
            The number of points for the representative grid in each dimension. By default (80,80,80)
        delta : float, optional
            The length added in all directions for the construction of the representative grid. By default 5
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    plotter = pyvista.Plotter()

    def create_mesh_AO(value,cutoff):
        AO_number = int(round(value)) - 1
        AO = AO_calculated[AO_number]
        AO = np.sign(AO)*AO**2
        plot_cube(plotter,coords,AO,opacity=opacity,factor=factor,cutoff=cutoff)

    coords,AO_calculated = rgmol.rectilinear_grid_reconstruction.reconstruct_AO(self,grid_points=grid_points,delta=delta)

    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)

    slider_function = _slider(create_mesh_AO,1,cutoff)

    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(lambda value: slider_function("AO",value), [1, len(AO_calculated)],value=1,title="Number", fmt="%1.0f")
    plotter.add_slider_widget(lambda value: slider_function("cutoff",value), [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.show(full_screen=False)
    return





def plot_MO(self,grid_points=(80,80,80),delta=5,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_MO(grid_points=(80,80,80),delta=5,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot the Molecular Orbitals of a molecule.
    Because no calculations are done, the Molecular Orbitals will be calculated on the representative grid.
    The representative grid is a cubic grid around the molecule.
    The delta defines the length to be added to the extremities of the position of the atoms.

    Parameters
    ----------
        grid_points : list of 3, optional
            The number of points for the representative grid in each dimension. By default (80,80,80)
        delta : float, optional
            The length added in all directions for the construction of the representative grid. By default 5
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    plotter = pyvista.Plotter()

    coords,MO_calculated = rgmol.rectilinear_grid_reconstruction.reconstruct_MO(self,grid_points=grid_points,delta=delta)


    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)

    def create_mesh_MO(value,cutoff):
        MO_number = int(round(value))
        MO = MO_calculated[MO_number-1]
        MO = np.sign(MO)*MO**2

        plot_cube(plotter,coords,MO,opacity=opacity,factor=factor,cutoff=cutoff)
        plotter.add_text(text=r"Energy = "+'{:3.3f} (a.u.)'.format(self.properties["MO_energy"][MO_number-1]),name="mo energy")
        print_occupancy(plotter,self.properties["MO_occupancy"],MO_number)

    slider_function = _slider(create_mesh_MO,1,cutoff)


    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(lambda value: slider_function("AO",value), [1, len(MO_calculated)],value=1,title="Number", fmt="%1.0f")
    plotter.add_slider_widget(lambda value: slider_function("cutoff",value), [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.show(full_screen=False)
    return



def plot_product_MO(self,grid_points=(80,80,80),delta=5,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_product_MO(grid_points=(80,80,80),delta=5,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot the Molecular Orbitals of a molecule on the left and the right.
    The Molecular Orbitals will be calculated on the grid that will be defined by the number of grid points and around the molecule.
    The delta defines the length to be added to the extremities of the position of the atoms.
    The center part represents the product of both sides which is calculated when pushing the button.

    Parameters
    ----------
        grid_points : list of 3, optional
            The number of points for the representative grid in each dimension. By default (80,80,80)
        delta : float, optional
            The length added in all directions for the construction of the representative grid. By default 3
        cutoff : float, optional
            The cutoff of the isodensity plot. By default .2
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    plotter = pyvista.Plotter(shape=(1,3),border=True)
    coords,AO_calculated = rgmol.rectilinear_grid_reconstruction.reconstruct_AO(self,grid_points=grid_points,delta=delta)
    MO_list = self.properties["MO_list"]

    if with_radius:
        plotter.subplot(0,0)
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)
        plotter.subplot(0,1)
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)
        plotter.subplot(0,2)
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)

    self.properties["MO_number_1"] = 1
    self.properties["MO_number_2"] = 1



    def create_mesh_MO_1(value):
        plotter.subplot(0,0)
        MO_number_1 = int(round(value))
        self.properties["MO_number_1"] = MO_number_1
        coords,MO = rgmol.rectilinear_grid_reconstruction.reconstruct_chosen_MO(self,MO_number_1-1,grid_points=grid_points,delta=delta)
        plot_cube(plotter,coords,MO,opacity=opacity,factor=factor,cutoff=cutoff,add_name="1")
        plotter.add_text(text=r"Energy = "+'{:3.3f} (a.u.)'.format(self.properties["MO_energy"][MO_number_1-1]),name="mo energy")
        print_occupancy(plotter,self.properties["MO_occupancy"],MO_number_1)


    def create_mesh_MO_2(value):
        plotter.subplot(0,2)
        MO_number_2 = int(round(value))
        self.properties["MO_number_2"] = MO_number_2
        coords,MO = rgmol.rectilinear_grid_reconstruction.reconstruct_chosen_MO(self,MO_number_2-1,grid_points=grid_points,delta=delta)
        plot_cube(plotter,coords,MO,opacity=opacity,factor=factor,cutoff=cutoff,add_name="2")
        plotter.add_text(text=r"Energy = "+'{:3.3f} (a.u.)'.format(self.properties["MO_energy"][MO_number_2-1]),name="mo energy")
        print_occupancy(plotter,self.properties["MO_occupancy"],MO_number_2)


    def calculate_product_MO(MO_ind_1,MO_ind_2):
        coords,MO_1 = rgmol.rectilinear_grid_reconstruction.reconstruct_chosen_MO(self,MO_ind_1-1,grid_points=grid_points,delta=delta)
        coords,MO_2 = rgmol.rectilinear_grid_reconstruction.reconstruct_chosen_MO(self,MO_ind_2-1,grid_points=grid_points,delta=delta)
        MO_prod = MO_1 * MO_2
        plot_cube(plotter,coords,MO_prod,opacity=opacity,factor=factor,cutoff=cutoff,add_name="prod")
        plotter.add_text(text=r"Product of {} and {}".format(MO_ind_1,MO_ind_2),name="mo prod")

    def button_product_MO(value):
        plotter.subplot(0,1)
        calculate_product_MO(self.properties["MO_number_1"],self.properties["MO_number_2"])




    plotter.subplot(0,0)
    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(create_mesh_MO_1, [1, len(MO_list)],value=1,title="Number", fmt="%1.0f")

    plotter.subplot(0,1)
    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_checkbox_button_widget(button_product_MO,size=80,color_off="blue")
    plotter.add_text(text="Calculate the product",name="calculate",position=(90.,20.))

    plotter.subplot(0,2)
    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(create_mesh_MO_2, [1, len(MO_list)],value=1,title="Number", fmt="%1.0f")

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.show(full_screen=False)
    return



def plot_electron_density(self,grid_points=(80,80,80),delta=5,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_electron_density(grid_points=(80,80,80),delta=5,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))


    Plot the Electron Density of a molecule.
    Because no calculations are truly done, the electron density will be calculated on the representative grid.
    The representative grid is a cubic grid around the molecule.
    The delta defines the length to be added to the extremities of the position of the atoms.
    The order of the Atomic Orbitals is defined in the molden file

    Parameters
    ----------
        grid_points : list of 3, optional
            The number of points for the representative grid in each dimension. By default (80,80,80)
        delta : float, optional
            The length added in all directions for the construction of the representative grid. By default 5
        cutoff : float, optional
            The cutoff of the isodensity plot. By default .2
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    plotter = pyvista.Plotter()
    coords,electron_density = rgmol.rectilinear_grid_reconstruction.reconstruct_electron_density(self,grid_points)

    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)

    def create_mesh_cube(value):
        plot_cube(plotter,coords,electron_density,opacity=opacity,factor=factor,cutoff=value)

    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(create_mesh_cube, [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)



def plot_fukui_function(self,mol_p=None,mol_m=None,fukui_type="0",grid_points=(80,80,80),delta=5,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_fukui_function(mol_p=None,mol_m=None,fukui_type="0",grid_points=(80,80,80),delta=5,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))


    Plots the fukui function desired.
    The fukui functions are calculated using finite differences of electron density.
    Thus, the molecule with one electron added (mol_p) or removed (mol_m) must be computed and added as argument to this function.
    Because no calculations are truly done, the fukui function will be calculated on the representative grid.
    The representative grid is a cubic grid around the molecule.
    The delta defines the length to be added to the extremities of the position of the atoms.
    The order of the Atomic Orbitals is defined in the molden file

    Parameters
    ----------
        mol_p : molecule, optional
            The molecule with an electron added. Needed for calculating the softness kernel with a fukui_type of "0" or "+"
        mol_n : molecule, optional
            The molecule with an electron removed. Needed for calculating the softness kernel with a fukui_type of "0" or "-"
        fukui_type : molecule, optional
            The type of fukui function used to calculate the softness kernel. The available types are "0", "+" or "-"
        grid_points : list of 3, optional
            The number of points for the representative grid in each dimension. By default (80,80,80)
        delta : float, optional
            The length added in all directions for the construction of the representative grid. By default 5
        cutoff : float, optional
            The cutoff of the isodensity plot. By default .2
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    plotter = pyvista.Plotter()

    coords,f0,fp,fm = rgmol.rectilinear_grid_reconstruction.reconstruct_fukui_function(self,mol_p=mol_p,mol_m=mol_m,grid_points=grid_points,delta=delta)

    if "0" in fukui_type:
        fukui = f0
    elif "+" in fukui_type or "p" in fukui_type:
        fukui = fp
    elif "-" in fukui_type or "m" in fukui_type:
        fukui = fm
    else:
        raise ValueError('This kind of fukui function does not exist. Please chose between 0, + or -')

    if type(fukui) is None:
        raise TypeError("This fukui function could not be computed, did you give mol_m or mol_p ?")


    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)

    def create_mesh_cube(value):
        plot_cube(plotter,coords,fukui,opacity=opacity,factor=factor,cutoff=value)

    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(create_mesh_cube, [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)




def plot_dipole_moment(self,opacity=0.8,factor=1,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_dipole_moment(opacity=0.8,factor=1,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))


    Plots the electric and magnetic dipole moments of a molecule.
    Data has to be extracted from an orca TDDFT calculation using :doc:`rgmol.extract_orca.extract_transition<../extract_orca>`.

    Parameters
    ----------
        opacity : float, optional
            The opacity of the arrow. By default .5
        factor : float, optional
            The factor by which the norm of the arrow is multiplied. By default 1
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    plotter = pyvista.Plotter()

    center_of_mass = self.properties["center_of_mass"]
    D = self.properties["D"]
    P = self.properties["P"]
    M = self.properties["M"]
    state_transition = self.properties["state_transition"]

    self.plot(plotter,factor=factor_radius,opacity=opacity_radius)

    def create_mesh_dipole_moment(value,mode_D,mode_P,mode_M):
        transition_number = int(round(value))
        directions = [[],[],[]]
        if mode_D: directions[0] = np.array(D[transition_number-1])
        if mode_P: directions[1] = np.array(P[transition_number-1])
        if mode_M: directions[2] = np.array(M[transition_number-1])
        compt = 1
        for direc in range(3):
            direction = directions[direc]
            if len(direction)>0:
                norm = np.sum(direction**2)**(1/2)
                dipole_moment = pyvista.Arrow(start=center_of_mass,direction=direction,scale=factor*norm)
                if direc==0:
                    plotter.add_mesh(dipole_moment,opacity=opacity,pbr=True,roughness=.5,metallic=.2,color="blue",name="dipole moment{}".format(direc))
                    plotter.add_text(text="Electric dipole\nNorm = "+'{:3.3f} (a.u.)'.format(norm),name="norm{}".format(direc),position=(0.05,0.8-.07*(2-direc)),viewport=True,font_size=int(15*plotter.window_size[1]/1000),color="blue")
                elif direc==1:
                    plotter.add_mesh(dipole_moment,opacity=opacity,pbr=True,roughness=.5,metallic=.2,color="green",name="dipole moment{}".format(direc))
                    plotter.add_text(text="Electric dipole velocity\nNorm = "+'{:3.3f} (a.u.)'.format(norm),name="norm{}".format(direc),position=(0.05,0.8-.07*(2-direc)),viewport=True,font_size=int(15*plotter.window_size[1]/1000),color="green")
                else:
                    plotter.add_mesh(dipole_moment,opacity=opacity,pbr=True,roughness=.5,metallic=.2,color="red",name="dipole moment{}".format(direc))
                    plotter.add_text(text="Magnetic dipole\nNorm = "+'{:3.3f} (a.u.)'.format(norm),name="norm{}".format(direc),position=(0.05,0.8-.07*(2-direc)),viewport=True,font_size=int(15*plotter.window_size[1]/1000),color="red")
                compt+=1
            else:
                plotter.remove_actor("norm{}".format(direc))
                plotter.remove_actor("dipole moment{}".format(direc))

            plotter.add_text(text=r"Transition :"+ state_transition[transition_number-1] ,name="transi num",position=(0.05,.9),viewport=True)
        plotter.add_text(text=r"Energy = "+'{:3.3f} (a.u.)'.format(self.properties["energy_transition_spectra"][transition_number-1]),name="transition energy")


    three_buttons = _three_buttons(create_mesh_dipole_moment,1,0,0,0)


    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.add_slider_widget(lambda value: three_buttons("value",value), [1, len(D)],value=1,title="Transition number", fmt="%1.0f")
    plotter.add_checkbox_button_widget(lambda value_d: three_buttons("d",value_d) ,size=80,position=(20.,150))
    plotter.add_checkbox_button_widget(lambda value_p: three_buttons("p",value_p) ,size=80,position=(20.,250),color_on="green")
    plotter.add_checkbox_button_widget(lambda value_m: three_buttons("m",value_m) ,size=80,position=(20.,350),color_on='red')



    plotter.show(full_screen=False)


def plot_transition_density(self,grid_points=(80,80,80),delta=5,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_transition_density(grid_points=(80,80,80),delta=5,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))


    Plot the Transition Densities of a molecule.
    All the AOs and the MOs will be calculated on the grid if they were not calculated.
    Because no calculations are truly done, the electron density will be calculated on the representative grid.
    The representative grid is a cubic grid around the molecule.
    The delta defines the length to be added to the extremities of the position of the atoms.
    The order of the Atomic Orbitals is defined in the molden file

    Parameters
    ----------
        grid_points : list of 3, optional
            The number of points for the representative grid in each dimension. By default (80,80,80)
        delta : float, optional
            The length added in all directions for the construction of the representative grid. By default 5
        cutoff : float, optional
            The cutoff of the isodensity plot. By default .2
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function

    Notes
    -----
    The transition densities are defined as :

    | :math:`\\rho_0^k = \\sum_i c_i (Occ_i(r) * Virt_i(r))`
    | With the sum being on all the transitions of the excitation, :math:`Occ_i(r)` and :math:`Virt_i(r)` being respectively the occupied and the virtual molecular orbitals considered in the transition, and :math:`c_i` the coefficient of the transition.

    Examples
    --------
    For a TD-DFT calculation output :

    STATE  1:  E=   0.148431 au      4.039 eV
       | 18a ->  20a  :     0.5000 (c= 0.7071)
       | 19a ->  21a  :     0.5000 (c= -0.7071)

    The transition density will be :
        | :math:`\\rho_0^1 =  c_1 (\\psi_{18}(r) * \\psi_{20}(r)) + c_2 (\\psi_{19}(r) * \\psi_{21}(r))`
        | with :math:`c_1 = 0.7071` and :math:`c_2 = -0.7071`
    """

    plotter = pyvista.Plotter()
    # coords,transition_density_list = rgmol.rectilinear_grid_reconstruction.reconstruct_transition_density(self,grid_points=grid_points,delta=delta)
    coords,MO_calculated = rgmol.rectilinear_grid_reconstruction.reconstruct_MO(self,grid_points=grid_points,delta=delta)


    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)


    def create_mesh_transition_density(value,cutoff):
        transition_number = int(round(value))
        coords,transition_density_calculated = rgmol.rectilinear_grid_reconstruction.reconstruct_chosen_transition_density(self,transition_number-1,grid_points=grid_points,delta=delta)
        plot_cube(plotter,coords,transition_density_calculated,opacity=opacity,factor=factor,cutoff=cutoff)


        plotter.add_text(text=r"Energy = "+'{:3.3f} (a.u.)'.format(self.properties["transition_energy"][transition_number-1]),name="transition energy")

    slider_function = _slider(create_mesh_transition_density,1,cutoff)


    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.add_slider_widget(lambda value: slider_function("AO",value), [1, len(self.properties["transition_list"])],value=1,title="Number", fmt="%1.0f")
    plotter.add_slider_widget(lambda value: slider_function("cutoff",value), [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))
    plotter.show(full_screen=False)







def plot_diagonalized_kernel(self,kernel,mol_p=None,mol_m=None,fukui_type="0",number_plotted_eigenvectors=100,try_reading=True,save=True ,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000),grid_points=(80,80,80),delta=5):
    """
    plot_diagonalized_kernel(kernel,mol_p=None,mol_m=None,fukui_type="0",number_plotted_eigenvectors=100,try_reading=True,save=True,plotting_method="isodensity",number_isodensities=10 ,opacity=0.8,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000),grid_points=(80,80,80),delta=5)

    Calculate and diagonalize a kernel.
    The available kernels are : "linear_response_function" and "softness_kernel"
    Only the eigenmodes are computed using a method that does not calculate the whole kernel.
    All the computations will be done on the atomic grids.

    The eigenmodes will also be computed on the atomic grids.
    One might consider using :doc:`plot_on_atomic_grid<plot_on_atomic_grid>` for visualization

    By default the eigenmodes will also be plotted on the representative grid using the calculations on the atomic grids.
    If one wants not to do the plot, putting number_plotted_eigenvectors to 0 will not produce any plot.

    For the softness kernel as it is calculated using the Parr-Berkowitz relation, the fukui functions need to be computed. For that, a calculation adding (mol_p) or removing an electron (mol_m) needs to be done with the same geometry.

    Parameters
    ----------
        kernel : str
            The kernel to be diagonalized and plotted
        mol_p : molecule, optional
            The molecule with an electron added. Needed for calculating the softness kernel with a fukui_type of "0" or "+"
        mol_m : molecule, optional
            The molecule with an electron removed. Needed for calculating the softness kernel with a fukui_type of "0" or "-"
        fukui_type : molecule, optional
            The type of fukui function used to calculate the softness kernel. The available types are "0", "+" or "-"
        number_plotted_eigenvectors : int, optional
            The Number of eigenvectors to be plotted. If 0, no plot will be produced. By default 100
        try_reading : bool, optional
            If the eigenmodes were calculated and saved before, try reading them from the rgmol folder. If False, the computation will be remade. True by default
        save : bool, optional
            The eigenmodes, eigenvalues and contributions of transition densities will be save in the folder rgmol that will be created where the molecule file is located
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot for the isodensity plotting method. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)
        grid_points : list of 3, optional
            The number of points for the representative grid in each dimension. By default (80,80,80)
        delta : float, optional
            The length added in all directions for the construction of the representative grid. By default 5

    Returns
    -------
        None
            The plotter should display when using this function and all the properties will be sorted inside the molecule.properties
    """
    did_read = 1
    if kernel == "linear_response_function":
        if "Reconstructed_linear_response_eigenvectors" in self.properties:
            print("############################################")
            print("# Eigenmodes already extracted or computed #")
            print("############################################")


        elif try_reading:
            try: self.read()
            except: pass

        if try_reading and not "Reconstructed_linear_response_eigenvectors" in self.properties:
            did_read = 0
            if not "contribution_linear_response_eigenvectors" in self.properties:
                print("########################################")
                print("# No previous calculations were found. #")
                print("#   The eigenmodes will be computed.   #")
                print("########################################")
                self.calculate_eigenmodes_linear_response_function()
            else:
                print("#########################################")
                print("#    Previous calculation found but     #")
                print("# the eigenmodes were not reconstructed #")
                print("#########################################")
        elif not try_reading:
            did_read = 0
            self.calculate_eigenmodes_linear_response_function()

        # eigenvectors = self.properties["linear_response_eigenvectors"]
        eigenvalues = self.properties["linear_response_eigenvalues"]
        contrib_eigenvectors = self.properties["contribution_linear_response_eigenvectors"]
        transition_list = self.properties["transition_list"]
        transition_factor_list = self.properties["transition_factor_list"]

        coords,eigenvectors = rgmol.rectilinear_grid_reconstruction.reconstruct_eigenvectors(self,kernel)

    elif kernel == "softness_kernel":
        self.calculate_hardness()
        if try_reading:
            try: self.read()
            except: pass

        if try_reading and not "Reconstructed_softness_kernel_eigenvectors" in self.properties:
            did_read = 0
            if not "contribution_softness_kernel_eigenvectors" in self.properties:
                print("########################################")
                print("# No previous calculations were found. #")
                print("#   The eigenmodes will be computed.   #")
                print("########################################")
                self.calculate_softness_kernel_eigenmodes(mol_p=mol_p,mol_m=mol_m,fukui_type=fukui_type)
            else:
                print("#########################################")
                print("#    Previous calculation found but     #")
                print("# the eigenmodes were not reconstructed #")
                print("#########################################")
        elif not try_reading:
            did_read = 0
            self.calculate_softness_kernel_eigenmodes(mol_p=mol_p,mol_m=mol_m,fukui_type=fukui_type)

        # eigenvectors = self.properties["softness_kernel_eigenvectors"]
        eigenvalues = self.properties["softness_kernel_eigenvalues"]
        contrib_eigenvectors = self.properties["contribution_softness_kernel_eigenvectors"]
        transition_list = self.properties["transition_list"] + [[[-1,-1]]]
        transition_factor_list = self.properties["transition_factor_list"] + [[[1]]]

        coords,eigenvectors = rgmol.rectilinear_grid_reconstruction.reconstruct_eigenvectors(self,kernel,mol_m=mol_m,mol_p=mol_p,grid_points=grid_points,delta=delta,fukui_type=fukui_type)


    else:
        raise TypeError("The kernel {} has not been implemented. The available kernels are : linear_response_function and softness_kernel".format(kernel))

    if save and not did_read:
        self.save(kernel)

    #Stop here if no eigenvectors are to be plotted
    if number_plotted_eigenvectors < 1:
        return


    eigenvectors = eigenvectors [:number_plotted_eigenvectors]
    eigenvalues = eigenvalues [:number_plotted_eigenvectors]


    plotter = pyvista.Plotter()
    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)


    # if plotting_method == "isodensity":
    def create_mesh_diagonalized_kernel(value,cutoff):
        vector_number = int(round(value))
        plot_cube(plotter,coords,eigenvectors[vector_number-1],opacity=opacity,factor=factor,cutoff=cutoff)
        plotter.add_text(text=r"eigenvalue = "+'{:3.3f} (a.u.)'.format(eigenvalues[vector_number-1]),name="eigenvalue")

        print_contribution_transition_density(plotter,self,kernel,vector_number,contrib_eigenvectors,transition_list,transition_factor_list)

    slider_function = _slider(create_mesh_diagonalized_kernel,1,cutoff)
    plotter.add_slider_widget(lambda value: slider_function("cutoff",value), [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))

    # elif plotting_method == "multiple isodensities":
    #     def create_mesh_diagonalized_kernel(value,cutoff):
    #         vector_number = int(round(value))
    #         plot_cube_multiple_isodensities(plotter,self.properties["voxel_origin"],self.properties["voxel_matrix"],eigenvectors[vector_number-1],factor=factor)
    #         plotter.add_text(text=r"eigenvalue = "+'{:3.3f} (a.u.)'.format(eigenvalues[vector_number-1]),name="eigenvalue")
    #
    #         print_contribution_transition_density(plotter,self,kernel,vector_number,contrib_eigenvectors,transition_list,transition_factor_list)


    slider_function = _slider(create_mesh_diagonalized_kernel,1,cutoff)
    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.add_slider_widget(lambda value: slider_function("AO",value), [1, len(eigenvectors)],value=1,title="Eigenvector", fmt="%1.0f")
    plotter.show(full_screen=False)




def plot_analysis_kernel(self,kernel,mol_p=None,mol_m=None,fukui_type="0",list_vectors=[1,2,3,4,5],try_reading_diagonalized=True,save_diagonalized=True,save=True,file_name="analysis"):
    """
    plot_analysis_kernel(kernel,mol_p=None,mol_m=None,fukui_type="0",list_vectors=[1,2,3,4,5],try_reading_diagonalized=True,save_diagonalized=True,save=1,file_name="analysis")

    Computes and plot some analysis of the eigenmodes of the linear_response_function or the softness_kernel.
    The analysis are : Percentage of occupied MO in the mode, percentage of virtual MO in the mode.
    For the softness kernel, the proportion of electron transferred in each mode.
    By default this will produce a file called analysis.png

    Parameters
    ----------
        kernel : str
            The kernel on which the analysis will be made. Either "linear_reponse_function" or "softness_kernel"
        mol_p : molecule, optional
            The molecule with an electron added. Needed for calculating the softness kernel with a fukui_type of "0" or "+"
        mol_m : molecule, optional
            The molecule with an electron removed. Needed for calculating the softness kernel with a fukui_type of "0" or "-"
        fukui_type : molecule, optional
            The type of fukui function used to calculate the softness kernel. The available types are "0", "+" or "-"
        list_vectors : list, optional
            The list of the vectors that will be decomposed on Occupied Virtual orbitals. Starting to 1 instead of 0. By default [1,2,3,4,5]
        try_reading_diagonalized : bool, optional
            If the eigenmodes will be read or not. By default True
        save_diagonalized : bool, optional
            If the eigenmodes will be saved or not. By default True
        save : bool, optional
            If the image is saved or showed. By default True
        file_name : str, optional
            The name of the file, the extension is not necessary as the default one is png. But if one wants to save in another format, putting the format will save the file in that format. By default "analysis"

    Returns
    -------
        None
            The image is either save or plotted
    """

    if kernel == "linear_response_function":
        if not "linear_response_eigenvectors" in self.list_properties():
            self.plot_diagonalized_kernel(kernel,number_plotted_eigenvectors=0,try_reading=try_reading_diagonalized,save=save_diagonalized)

    elif kernel == "softness_kernel":
        if not "softness_kernel_eigenvectors" in self.list_properties():
            self.plot_diagonalized_kernel(kernel,number_plotted_eigenvectors=0,mol_p=mol_p,mol_m=mol_m,fukui_type=fukui_type,try_reading=try_reading_diagonalized,save=save_diagonalized)

    else:
        raise ValueError("The kernel {} has not been implemented. The available kernels are : linear_response_function and softness_kernel".format(kernel))

    self.analysis_eigenmodes(kernel=kernel,list_vectors=list_vectors)

    from_occ = np.array(self.properties["from_occ"])
    to_virt = np.array(self.properties["to_virt"])

    colors = ["red","blue","green","darkred","cyan","lime","magenta","teal","purple","darkorange"]

    fig=plt.figure(figsize=(15,7),dpi=200)
    plt.rcParams.update({'font.size': 15})
    plt.rcParams['svg.fonttype'] = 'none'

    if kernel == "linear_response_function":
        plt.subplot(1,2,1)
        for index in range(len(from_occ)):
            plt.plot(from_occ[index],"o-",color=colors[index],label=list_vectors[index])

    elif kernel == "softness_kernel":
        plt.subplot(2,2,1)
        for index in range(len(from_occ)):
            plt.plot(from_occ[index,:-1],"o-",color=colors[index],label=list_vectors[index])
            plt.plot(len(from_occ[0])-1,from_occ[index,-1],"o",color=colors[index])
        plt.xticks(ticks=np.arange(len(from_occ[0])),label=[str(k) for k in range(len(from_occ[0])-1)] + ["f"])

    plt.xlabel("Occupied MO index",size=17)
    plt.ylabel("Coefficient",size=17)
    plt.legend()

    n_occ = len(from_occ[0])
    n_virt = len(to_virt[0])

    if kernel == "linear_response_function":
        plt.subplot(1,2,2)
    elif kernel == "softness_kernel":
        plt.subplot(2,2,2)
        n_occ-=1
    for index in range(len(to_virt)):
        plt.plot(np.arange(n_occ,n_virt),to_virt[index][n_occ:],"o-",color=colors[index],label=list_vectors[index])
    plt.xlabel("Virtual MO index",size=17)
    plt.ylabel("Coefficient",size=17)

    if kernel == "softness_kernel":
        plt.subplot(2,1,2)
        fukui_decomposition = self.properties["fukui_decomposition"]
        plt.plot(fukui_decomposition,"o-r")
        plt.xlabel("Eigenmode Index",size=17)
        plt.ylabel("Electron Transfer Coefficient",size=17)
        # plt.ylim(0,1)


    plt.tight_layout()
    if save:
        plt.savefig(file_name)
    else: plt.show()






molecule.plot = plot
molecule.plot_vector = plot_vector
molecule.plot_radius = plot_radius
molecule.plot_property = plot_property
molecule.plot_condensed_kernel = plot_condensed_kernel
molecule.plot_diagonalized_condensed_kernel = plot_diagonalized_condensed_kernel
molecule.plot_isodensity = plot_isodensity
molecule.plot_multiple_isodensities = plot_multiple_isodensities
molecule.plot_AO = plot_AO
molecule.plot_MO = plot_MO
molecule.plot_product_MO = plot_product_MO
molecule.plot_electron_density = plot_electron_density
molecule.plot_dipole_moment = plot_dipole_moment
molecule.plot_transition_density = plot_transition_density
molecule.plot_fukui_function = plot_fukui_function
molecule.plot_diagonalized_kernel = plot_diagonalized_kernel
molecule.plot_analysis_kernel = plot_analysis_kernel





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script adds functions, and methods to the molecule objects.
These methods allow the calculation of various chemical properties such as AO, MO, Transition densities or linear response function.
"""

import numpy as np
import scipy as sp
from rgmol.objects import *
from rgmol.threading_functions import *
import scipy as sp
from rgmol.dicts import *


class mol_grids:
    """
    mol_grids(grids)

    Constructs a list of multiple atomic grids for a molecule

    Parameters
    ----------
        grids : list, optional
            List containing all the grids

    Returns
    -------
        mol_grids
            mol_grids object

    Attributes
    ----------
        grids : list
            The list of all the grids
        centers : list
            The list of the centers
    """
    def __init__(self,grids):
        """
        mol_grids(grids)

        Constructs a list of multiple atomic grids for a molecule

        Parameters
        ----------
            grids : list, optional
                List containing all the grids

        Returns
        -------
            mol_grids
                mol_grids object

        Attributes
        ----------
            grids : list
                The list of all the grids
            centers : list
                The list of the centers
            dV : array
                The dV of all grids
        """
        self.grids = grids
        grids_centers = []

        for grid_i in grids:
            grids_centers.append(grid_i.center)

        self.grids_centers = np.array(grids_centers)


    def integrate(self,arr):
        """
        integrate(arr)

        Integrates an array on all the atomic grid

        Parameters
        ----------
            arr : ndarray
                The array to be integrated

        Returns
        -------
            Int
                the integral
        """

        Int = 0
        for grid,grid_index in zip(self.grids,range(len(self.grids))):
            Int += grid.integrate(arr[grid_index])
        return Int


    def integrate_product(self,arr,arr_2):
        """
        integrate_product(arr,arr_2)

        Integrates the product of two arrays on all the atomic grid.
        This method should be used as it is around 3 times faster as using integrate with the product as an argument

        Parameters
        ----------
            arr : ndarray
                The array to be multiplied and integrated
            arr_2 : ndarray
                The array to be multiplied and integrated

        Returns
        -------
            Int
                the integral
        """

        Int = 0
        for grid,grid_index in zip(self.grids,range(len(self.grids))):
            Int += grid.integrate_product(arr[grid_index],arr_2[grid_index])
        return Int

    def get_dV(self):
        """
        get_dV

        Gets the integration elements for all grids in the format : N_grids,N_r,N_ang

        Parameters
        ----------
            None

        Returns
        -------
            dV
                The integration elements
        """
        N_r,N_ang = np.shape(self.grids[0].dV)
        dV = np.zeros((len(self.grids),N_r,N_ang))
        for grid_i,index_grid in zip(self.grids,range(len(self.grids))):
            dV[index_grid] = grid_i.dV * grid_i.wn

        return dV

class grid:
    """
    grid(coords,center,grid_type,dV,coords_gc=None)

    Constructs an atomic grid around a center.
    For contructing the grid properly around an atom, one should use create_atomic_grid

    Parameters
    ----------
        coords : ndarray
            Array containing the spherical coordinates of each point of the grid
        center : ndarray
            The center of the grid
        grid_type : str
            The type of grid, should be atomic as cubic are deprecated
        dV : array
            The dV of each point used for integration
        coords_gc : array, optional
            The coordinates of the gauss-chebychev integration.
            This represents the coordinates between -1 and 1
    Returns
    -------
        grid
            grid object

    Attributes
    ----------
        coords : ndarray
        center : ndarray
        grid_type : str
        dV : array
        coords_gc : array
        number_points : int
        xyz_coords : array
    """
    def __init__(self,coords,center,grid_type,dV,coords_gc=None):
        """
        grid(coords,center,grid_type,dV,coords_gc=None)

        Constructs an atomic grid around a center.
        For contructing the grid properly around an atom, one should use create_atomic_grid

        Parameters
        ----------
            coords : ndarray
                Array containing the spherical coordinates of each point of the grid
            center : ndarray
                The center of the grid
            grid_type : str
                The type of grid, should be atomic as cubic are deprecated
            dV : array
                The dV of each point used for integration
            coords_gc : array, optional
                The coordinates of the gauss-chebychev integration.
                This represents the coordinates between -1 and 1
        Returns
        -------
            grid
                grid object

        Attributes
        ----------
            coords : ndarray
            center : ndarray
            grid_type : str
            dV : array
            coords_gc : array
            number_points : int
            xyz_coords : array
        """

        self.grid_type = grid_type
        self.center = center
        self.coords = coords
        self.number_points = np.prod(np.shape(coords)[1:])
        self.coords_gc = coords_gc
        self.dV = dV

        if grid_type.lower() == "cubic":
            self.xyz_coords = coords

        elif grid_type.lower() == "atomic":
            r,Theta,Phi = coords
            cos_theta = np.cos(Theta)
            sin_theta = np.sin(Theta)
            cos_phi = np.cos(Phi)
            sin_phi = np.sin(Phi)

            x = r * sin_theta * cos_phi + self.center[0]
            y = r * sin_theta * sin_phi + self.center[1]
            z = r * cos_theta + self.center[2]

            self.xyz_coords = np.array([x,y,z])

        else:
            raise TypeError("This type of grid {} has not been implemented. Please use cubic or atomic.".format(grid_type))



    def integrate(self,arr):
        """
        integrate(arr)

        Integrates an array on the atomic grid

        Parameters
        ----------
            arr : ndarray
                The array to be integrated

        Returns
        -------
            Int
                the integral
        """

        Coords = self.coords
        wn = self.wn
        dV = self.dV

        dV = wn * dV

        Int = np.einsum("ij,ij->",arr,dV)

        return Int


    def integrate_product(self,arr,arr_2):
        """
        integrate_product(arr,arr_2)

        Integrates the product of two arrays on the atomic grid.
        This method should be used as it is around 3 times faster as using integrate with the product as an argument

        Parameters
        ----------
            arr : ndarray
                The array to be multiplied and integrated
            arr_2 : ndarray
                The array to be multiplied and integrated

        Returns
        -------
            Int
                the integral
        """

        Coords = self.coords
        wn = self.wn
        dV = self.dV

        dV = wn * dV

        Int = np.einsum("ij,ij,ij->",arr,arr_2,dV)

        return Int

#################
## Cubic grids ##
#################

def create_cubic_grid_from_molecule(mol,grid_points,delta=5):
    """
    create_cubic_grid_from_molecule(mol,grid_points,delta=5)

    Creates a cubic voxel from the position of atoms in a molecule

    Parameters
    ----------
        mol : molecule_object
        grid_points : list of 3
            the number of points for each coordinates
        delta : float, optional
            the length added on all directiosn to the box containing all atomic centers

    Returns
    -------
        voxel_origin : list of 3
            the origin of the voxel
        voxel_matrix : 2d list
            the matrix of the voxel
    """

    mol.properties["grid_points"] = grid_points
    mol.properties["delta"] = delta

    list_pos = mol.list_property("pos")
    nx,ny,nz = grid_points
    xmin,ymin,zmin = np.min(list_pos,axis=0)
    xmax,ymax,zmax = np.max(list_pos,axis=0)

    voxel_origin = [xmin-delta,ymin-delta,zmin-delta]

    x_step = (xmax - xmin + 2*delta)/nx
    y_step = (ymax - ymin + 2*delta)/ny
    z_step = (zmax - zmin + 2*delta)/nz

    voxel_matrix = [[x_step,0,0],[0,y_step,0],[0,0,z_step]]
    r,c = create_coordinates_from_cubic_grid(grid_points,voxel_origin,voxel_matrix)

    return r,c

def create_coordinates_from_cubic_grid(grid_points,voxel_origin,voxel_matrix):
    """
    create_coordinates_from_cubic_grid(grid_points,voxel_origin,voxel_matrix)

    Creates coordinates from a cubic grid

    Parameters
    ----------
        grid_points : list of 3
        voxel_origin : list of 3
        voxel_matrix : 2d list

    Return
    ------
        r : ndarray corresponding to the coordinates
    """

    nx,ny,nz = grid_points

    voxel_end = [0,0,0]

    voxel_end[0] = voxel_origin[0] + voxel_matrix[0][0]*nx
    voxel_end[1] = voxel_origin[1] + voxel_matrix[1][1]*ny
    voxel_end[2] = voxel_origin[2] + voxel_matrix[2][2]*nz


    x = np.linspace(voxel_origin[0],voxel_end[0],nx,endpoint=True)
    y = np.linspace(voxel_origin[1],voxel_end[1],ny,endpoint=True)
    z = np.linspace(voxel_origin[2],voxel_end[2],nz,endpoint=True)

    r = x.reshape((1,nx,1,1))*np.array([1.,0,0]).reshape((3,1,1,1)) + \
        y.reshape((1,1,ny,1))*np.array([0,1.,0]).reshape((3,1,1,1)) + \
        z.reshape((1,1,1,nz))*np.array([0,0,1.]).reshape((3,1,1,1))

    return r,[x,y,z]


def create_rectilinear_grid_from_molecule(mol,grid_points=(80,80,80),delta=7):
    """
    create_cubic_grid_from_molecule(mol,grid_points=(80,80,80),delta=7)

    Creates a rectilinear grid from the position of atoms in a molecule

    Parameters
    ----------
        mol : molecule_object
        grid_points : list of 3
            the number of points for each coordinates
        delta : float, optional
            the length added on all directiosn to the box containing all atomic centers

    Returns
    -------
        voxel_origin : list of 3
            the origin of the voxel
        voxel_matrix : 2d list
            the matrix of the voxel
    """

    mol.properties["grid_points"] = grid_points
    mol.properties["delta"] = delta

    list_pos = mol.list_property("pos")
    nx,ny,nz = grid_points
    xmin,ymin,zmin = np.min(list_pos,axis=0)
    xmax,ymax,zmax = np.max(list_pos,axis=0)

    N_atoms = len(mol.atoms)

    nx,ny,nz = grid_points
    zeta=1
    alpha=0.6

    list_x = []
    list_y = []
    list_z = []

    nx,ny,nz = nx//2,ny//2,nz//2
    #Gauss Chebyshev Type 2
    mx = np.cos((np.arange(nx)+1)/(nx+1) *np.pi)
    mx = mx - (mx==1)*0.0001 #Convert x=1 to x=0.999

    x = zeta / np.log(2) * np.log( (2)/ (1-mx) ) * (1+mx)**(alpha) #M4
    x_tot = np.append(-x,x[::-1])

    my = np.cos((np.arange(ny)+1)/(ny+1) *np.pi)
    my = my - (my==1)*0.0001 #Convert x=1 to x=0.999
    y = zeta / np.log(2) * np.log( (2)/ (1-my) ) * (1+my)**(alpha) #M4
    y_tot = np.append(-y,y[::-1])

    mz = np.cos((np.arange(nz)+1)/(nz+1) *np.pi)
    mz = mz - (mz==1)*0.0001 #Convert x=1 to x=0.999
    z = zeta / np.log(2) * np.log( (2)/ (1-mz) ) * (1+mz)**(alpha) #M4
    z_tot = np.append(-z,z[::-1])

    r = x_tot.reshape((1,nx*2,1,1))*np.array([1.,0,0]).reshape((3,1,1,1)) + \
        y_tot.reshape((1,1,ny*2,1))*np.array([0,1.,0]).reshape((3,1,1,1)) + \
        z_tot.reshape((1,1,1,nz*2))*np.array([0,0,1.]).reshape((3,1,1,1))

    # d = dx_tot.reshape((1,nx*2,1,1))*np.array([1.,0,0]).reshape((3,1,1,1)) + \
    #     dy_tot.reshape((1,1,ny*2,1))*np.array([0,1.,0]).reshape((3,1,1,1)) + \
    #     dz_tot.reshape((1,1,1,nz*2))*np.array([0,0,1.]).reshape((3,1,1,1))

    return r,[x_tot,y_tot,z_tot]


##################
## Atomic grids ##
##################

def create_grid_from_mol(mol,N_r=None,d_leb=None,zeta_list=None,alpha_list=None):
    """
    create_grid_from_mol(mol,N_r=None,d_leb=None,zeta_list=None,alpha_list=None)

    Creates atomic grids for all the atoms in the molecule.
    For more details on the atomic grids, see :doc:`create_atomic_grid<create_atomic_grid>`

    Parameters
    ----------
        mol : molecule
            The molecule for which the grids will be computed
        N_r : list, optional
            The number of radial points for each atomic grid
            If not provided, an already optimized value will be used
        d_leb : list, optional
            The order of Lebedev quadrature used for each atomic grid
            If not provided, an already optimized value will be used
        zeta_list : list, optional
            The zeta parameter, more info on :doc:`create_atomic_grid<create_atomic_grid>`
            If not provided, an already optimized value will be used
        alpha_list : list, optional
            The alpha parameter, more info on :doc:`create_atomic_grid<create_atomic_grid>`
            If not provided, an already optimized value will be used

    Returns
    -------
        None
            The molecule attribute mol_grids will contain the result
    """

    if mol.mol_grids:
        return

    list_grids = []

    N_r_list = []
    d_leb_list = []
    for atom in mol.atoms:
        if not N_r:
            N_r = dict_N_r[atom.name]
        if not d_leb:
            d_leb = dict_d_leb[atom.name]
        N_r_list.append(N_r)
        d_leb_list.append(d_leb)
    d_leb = np.max(d_leb_list)
    N_r_list = np.max(N_r_list)

    for atom,index_atom in zip(mol.atoms,range(len(mol.atoms))):
        zeta, alpha = None, None
        if alpha_list: alpha = alpha_list[index_atom]
        if zeta_list: zeta = zeta_list[index_atom]

        list_grids.append(create_atomic_grid(atom,N_r=N_r,d_leb=d_leb,zeta=zeta,alpha=alpha))



    m_grids = mol_grids(list_grids)
    voronoi_becke(m_grids)
    mol.mol_grids = m_grids





def create_atomic_grid(atom,N_r=None,d_leb=None,zeta=None,alpha=None):
    """
    create_atomic_grid(atom,N_r=None,d_leb=None,zeta=None,alpha=None)

    Creates an atomic grid
    It follows the Treutler and Ahlrichs M4 formula [1]_ for the radial part.
    And the Lebedev grid [2]_ for the angular part.
    If the parameters of the grid are not provided, already optimized ones will be used.


    Parameters
    ----------
        atom : atom
            the atom at the center of the grid
        N_r : int, optional
            the number of points for the radial part
        d_leb : int, optional
            the degree of the Lebedev quadrature
        zeta : float, optional
            the zeta parameter of the radial function M4
        alpha : float, optional
            the alpha parameter of the radial function M4

    Return
    ------
        grid

    Notes
    -----

        The formula M4 is the following :
        :math:`r = \\frac{\\zeta}{ln(2)}(1+x)^{\\alpha}ln\\left(\\frac{2}{1-x}`

    References
    ----------

        .. [1] Treutler, O.; Ahlrichs, R.The Journal of Chemical Physics1995,102,346–354.
        .. [2] Lebedev, V. I.; Laikov, D. N. InDoklady Mathematics, 1999; Vol. 59,pp 477–481.
    """

    if not zeta:
        zeta = dict_zeta[atom.name]
    if not alpha:
        alpha = dict_alpha[atom.name]

    coords_lebedev_xyz,weight_lebedev = sp.integrate.lebedev_rule(d_leb)
    X,Y,Z = coords_lebedev_xyz
    N_ang = len(X)

    X = X + (X==0)*1e-13 #remove dividing by 0
    R = 1 #Radius of the sphere

    #Convert to spheric coordinates
    Theta = np.arccos(Z/R)
    Phi = np.sign(Y) * np.arccos(X/(X**2+Y**2)**(1/2))
    Phi = np.arctan2(Y,X)


    #Gauss Chebyshev Type 2
    x = np.cos((np.arange(N_r)+1)/(N_r+1) *np.pi)
    x = x - (x==1)*0.0001 #Convert x=1 to x=0.999

    # r = R*(1+x)/(1-x) #Becke Transform
    # r = zeta / np.log(2) * np.log( (2)/ (1-x) ) #M3
    r = zeta / np.log(2) * np.log( (2)/ (1-x) ) * (1+x)**(alpha) #M4

    r = r.reshape((N_r,1))
    x = x.reshape((N_r,1))

    weight_lebedev = weight_lebedev.reshape((1,N_ang))

    coords = r.reshape((1,N_r,1))*np.array([1.,0,0]).reshape((3,1,1)) + \
             Theta.reshape((1,1,N_ang))*np.array([0,1.,0]).reshape((3,1,1)) + \
             Phi.reshape((1,1,N_ang))*np.array([0,0,1.]).reshape((3,1,1))

    # dr = 2/(1-x)**2 / (1-x**2)**(1/2) #Becke transform dr
    # dr = zeta/np.log(2) * (1/(1-x)) / (1-x**2)**(1/2) #M3 dr
    dr = zeta/np.log(2) * (1+x)**(alpha) * (alpha / (1+x) * np.log((2)/(1-x)) + 1/(1-x) ) / (1-x**2)**(1/2) #M4 dr
    w_gc = (np.pi/(N_r+1) * np.sin((np.arange(N_r)+1)/(N_r+1)*np.pi)**2).reshape((N_r,1)) #Weight Gauss Chebyshev Type 2

    dV = dr * w_gc * weight_lebedev * r * r
    grid_i = grid(coords,atom.pos,"atomic",coords_gc=x,dV=dV)

    return grid_i



def voronoi_becke(mol_grids):
    """
    voronoi_becke(mol_grids)

    This functions uses the Voronoi-Becke scheme for the partition of space [1]_
    This allows to divide the whole space in multiple atomic grids

    Parameters
    ----------
        mol_grids : mol_grids
            All the grids

    Returns
    -------
        None
            The weights are automatically attributed to each grid inside mol_grids


    References
    ----------
        .. [1] Becke, A. D.The Journal of Chemical Physics1988,88, 2547–2553.
    """

    def function_cutoff(mu):
        return 3/2*mu - 1/2*mu**3

    num_grids = len(mol_grids.grids)
    grids_centers = mol_grids.grids_centers


    for grid_i,grid_i_index in zip(mol_grids.grids,range(num_grids)):
        #Dimensions : grid_i,grid_j,rad,ang,coordinates
        number_points = grid_i.number_points
        n_r,n_l = np.shape(grid_i.xyz_coords)[1:]
        xyz_coords = grid_i.xyz_coords.transpose((1,2,0)).reshape((1,1,n_r,n_l,3))

        grids_centers_i = grids_centers.reshape((num_grids,1,1,1,3))
        grids_centers_j = grids_centers.reshape((1,num_grids,1,1,3))
        R_ij = np.linalg.norm(grids_centers_i-grids_centers_j,axis=4)
        R_ij = R_ij + (R_ij==0)*1 #Remove diagonal part

        r_i = np.linalg.norm(xyz_coords - grids_centers_i,axis=4)
        r_j = np.linalg.norm(xyz_coords - grids_centers_j,axis=4)
        Mu_ij = (r_i - r_j) / R_ij

        s3 = 1/2 * (1-function_cutoff(function_cutoff(function_cutoff(Mu_ij))))
        # s3 = Mu_ij<0 #voronoi scheme

        Pi = np.prod(s3,axis=1)
        wn = Pi[grid_i_index] / np.sum(Pi,axis=0)
        grid_i.wn = wn







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script adds functions, and methods to the molecule objects.
These methods allow the calculation of various chemical properties such as AO, MO, Transition densities or linear response function.
"""

import time
import numpy as np
import scipy as sp
import rgmol
from rgmol.objects import *
from rgmol.threading_functions import *


def calculate_fukui_function(self,mol_p=None,mol_m=None):
    """
    calculate_fukui_function(mol_p=None,mol_m=None)

    Calculates the fukui function using finite differences between electron density.
    If mol_p is provided, f+ will be computed.
    If mol_m is provided, f- will be computed.
    If both are provided, f+, f- and f0 will be computed.
    The fukui functions are automatically added to the molecule.properties with the keys : f0 f+ and f-

    Parameters
    ----------
    mol_p : molecule, optional
        A molecule with the same atoms and positions, but with an extra electron. By default None
    mol_m : molecule, optional
        A molecule with the same atoms and positions, but with an electron less. By default None

    Returns
    -------
    f0
        the f0 fukui function
    fp
        the f+ fukui function
    fm
        the f- fukui function


    Notes
    -----

    :math:`f^+ = \\rho(N+1)-\\rho(N)`
    :math:`f^- = \\rho(N)-\\rho(N-1)`
    :math:`f^0 = (f^+ + f^-)/2`
    """

    if not(mol_p) and not(mol_m):
        raise TypeError("Need at least the molecule with N+1 or N-1 number of electrons")

    if not "electron_density" in self.properties:
        self.calculate_electron_density()
    rhoN = self.properties["electron_density"]

    if mol_p:
        #calculate f+
        if not "electron_density" in mol_p.properties:
            mol_p.calculate_electron_density()

        rhoNp1 = mol_p.properties["electron_density"]

        fp = rhoNp1 - rhoN
    else:
        fp = None

    if mol_m:
        #calculate f-
        if not "electron_density" in mol_m.properties:
            mol_m.calculate_electron_density()

        rhoNm1 = mol_m.properties["electron_density"]

        fm = rhoN - rhoNm1

    else:
        fm = None

    if mol_p and mol_m:
        f0 = (fp+fm)/2
    else: f0 = None

    self.properties["f0"] = f0
    self.properties["f+"] = fp
    self.properties["f-"] = fm

    return f0,fp,fm


def calculate_hardness(self):
    """
    calculate_hardness()

    This method calculates the global hardness using Koopmans theorem.
    The hardness is therefore just the difference LUMO - HOMO
    The hardness is directly added into the molecule.properties with the key "hardness"

    Parameters
    ----------
    None

    Returns
    -------
    hardness
        the hardness of the molecule

    """

    if not "MO_energy" in self.properties:
        raise TypeError("The Energy of the MO has not been found. Cannot calculate the hardness")

    MO_occupancy = self.properties["MO_occupancy"]
    MO_energy =  self.properties["MO_energy"]

    LUMO = np.argmin(MO_occupancy)
    HOMO = LUMO - 1

    hardness = MO_energy[LUMO] - MO_energy[HOMO]
    self.properties["hardness"] = hardness
    return hardness


def calculate_eigenmodes_linear_response_function(self):
    """
    calculate_eigenmodes_linear_response_function()

    Calculates the linear response function from the transition densities.
    This method does not calculate directly the linear response, but only the eigenmodes.
    The mathematics behind this function will soon be available somewhere...

    Parameters
    ----------
        None
            All the data should be inside the molecule. Needs TD-DFT and molden properties

    Returns
    -------
    linear_response_eigenvalues
        the eigenvalues of the linear response function
    linear_response_eigenvectors
        the eigenvectors of the linear response function

    Notes
    -----
    The linear response function kernel can be computed as :

    :math:`\\chi(r,r') = -2\\sum_{k\\neq0} \\frac{\\rho_0^k(r) \\rho_0^k(r')}{E_k-E_0}`

    With :math:`\\rho_0^k` the transition density, and :math:`E_k` the energy of the transition k.

    Therefore, the molecule needs the transition properties that can be extracted from a TD-DFT calculation, and the MO extracted from a molden file. More details can be found :doc:`here<../tuto/orbitals>`.
    """

    if not "transition_density_list" in self.properties:
        self.calculate_transition_density()

    #Dummy implementation for now
    transition_density_list = self.properties["transition_density_list"]
    transition_energy = np.array(self.properties['transition_energy'])
    N_grids,N_trans,N_r,N_ang = np.shape(transition_density_list)

    nprocs = rgmol.nprocs

    print("#################################")
    print("# Calculating Eigenmodes of LRF #")
    print("#################################")
    time_before_calc = time.time()

    if nprocs >1:
        transition_matrix = calculate_overlap_matrix_multithread(self,transition_density_list,nprocs)
    else:
        transition_matrix = np.zeros((N_trans,N_trans))
        for first_transition in range(N_trans):
            for second_transition in range(first_transition+1):
                overlap_integral = self.mol_grids.integrate_product(transition_density_list[:,first_transition],transition_density_list[:,second_transition])

                transition_matrix[first_transition,second_transition] = overlap_integral
                transition_matrix[second_transition,first_transition] = overlap_integral

    #linear response matrix in transition densities basis
    #following dimensions : i,j,k
    # LR_matrix_in_TDB = np.einsum("k,ik,kj->ij",-2/transition_energy,transition_matrix,transition_matrix)

    # eigenvalues, eigenvectors = np.linalg.eigh(diag_matrix)
    # eigenvalues, eigenvectors = sp.linalg.eigh(LR_matrix_in_TDB,transition_matrix)
    # eigenvectors = eigenvectors.transpose()


    D = np.diag(-2/transition_energy).dot(transition_matrix)
    eigenvalues,eigenvectors = np.linalg.eig(D)
    eigenvectors = eigenvectors.transpose()


    # eigvalB,eigvecB = np.linalg.eigh(transition_matrix)
    # eigvecBtilde = eigvecB * eigvalB**(-1/2)
    # Atilde = eigvecBtilde.transpose() @ LR_matrix_in_TDB @ eigvecBtilde
    # eigenvalues, eigvecA = np.linalg.eigh(Atilde)
    # eigenvectors = eigvecBtilde @ eigvecA

    # eigenvectors = eigenvectors.transpose()

    if nprocs >1:
        reconstructed_eigenvectors, reconstructed_eigenvector_norms = multithreading_reconstruct_eigenvectors_atomic_grids(self,transition_density_list,eigenvectors,nprocs)
    else:
        reconstructed_eigenvectors = []
        reconstructed_eigenvector_norms = []
        for eigenvector in eigenvectors:
            reconstructed_eigenvector = np.einsum("j,ijkl->ikl",eigenvector,transition_density_list)
            #all eigenvectors are defined at a scalar and the square should integrate to 1
            reconstructed_eigenvector_norm = self.mol_grids.integrate_product(reconstructed_eigenvector,reconstructed_eigenvector)
            reconstructed_eigenvector_norms.append(reconstructed_eigenvector_norm)
            reconstructed_eigenvectors.append(reconstructed_eigenvector/reconstructed_eigenvector_norm**(1/2))

    eigenvectors = np.einsum("ij,i->ij",eigenvectors,1/np.array(reconstructed_eigenvector_norms)**(1/2))
    reconstructed_eigenvectors = np.array(reconstructed_eigenvectors)

    eig_sort = np.argsort(eigenvalues)
    eigenvalues = eigenvalues[eig_sort]
    eigenvectors = eigenvectors[eig_sort]
    reconstructed_eigenvectors = reconstructed_eigenvectors[eig_sort]


    self.properties["linear_response_eigenvalues"] = eigenvalues
    self.properties["contribution_linear_response_eigenvectors"] = np.array(eigenvectors)
    self.properties["linear_response_eigenvectors"] = reconstructed_eigenvectors

    time_taken = time.time() - time_before_calc

    print("##########################################")
    print("# Finished calculating eigenmodes of LRF #")
    print("# in {:3.3f} s #".format(time_taken))
    print("##########################################")
    return eigenvalues, np.array(eigenvectors)


def calculate_softness_kernel_eigenmodes(self,fukui_type="0",mol_p=None,mol_m=None):
    """
    calculate_softness_kernel_eigenmodes(fukui_type="0",mol_p=None,mol_m=None)

    Calculates the softness kernel from the transition densities and the fukui function using the Parr-Berkowitz relation.
    For that, a calculation adding (mol_p) or removing an electron (mol_m) needs to be done with the same geometry.
    This method does not calculate directly the softness kernel, but only the eigenmodes.
    The mathematics behind this function will soon be available somewhere...

    Parameters
    ----------
    mol_p : molecule, optional
        The molecule with an electron added. Needed for calculating the softness kernel with a fukui_type of "0" or "+"
    mol_n : molecule, optional
        The molecule with an electron removed. Needed for calculating the softness kernel with a fukui_type of "0" or "-"
    fukui_type : molecule, optional
        The type of fukui function used to calculate the softness kernel. The available types are "0", "+" or "-"

    Returns
    -------

    linear_response_eigenvalues
        the eigenvalues of the linear response function
    linear_response_eigenvectors
        the eigenvectors of the linear response function

    Notes
    -----
    The linear response function kernel can be computed as :

    :math:`\\chi(r,r') = -2\\sum_{k\\neq0} \\frac{\\rho_0^k(r) \\rho_0^k(r')}{E_k-E_0}`

    With :math:`\\rho_0^k` the transition density, and :math:`E_k` the energy of the transition k.

    Therefore, the molecule needs the transition properties that can be extracted from a TD-DFT calculation, and the MO extracted from a molden file. More details can be found :doc:`here<../tuto/orbitals>`.
    """


    if not "transition_density_list" in self.properties:
        self.calculate_transition_density()

    transition_density_list = self.properties["transition_density_list"]
    transition_energy = np.array(self.properties['transition_energy'])

    N_grids,N_trans,N_r,N_ang = np.shape(transition_density_list)

    if not fukui_type in self.properties:
        self.calculate_fukui_function(mol_p=mol_p,mol_m=mol_m)

    if "0" in fukui_type:
        fukui = self.properties["f0"]
    elif "+" in fukui_type or "p" in fukui_type:
        fukui = self.properties["f+"]
    elif "-" in fukui_type or "m" in fukui_type:
        fukui = self.properties["f-"]

    if not 'hardness' in self.properties:
        self.calculate_hardness()
    hardness = self.properties["hardness"]


    nprocs = rgmol.nprocs

    print("#############################################")
    print("# Calculating Eigenmodes of Softness Kernel #")
    print("#############################################")


    time_before_calc = time.time()
    fukui_reshaped = fukui.reshape((N_grids,1,N_r,N_ang))
    basis = np.append(transition_density_list,fukui_reshaped,axis=1) #Append the fukui function
    factor = np.array((transition_energy/2).tolist() + [hardness])

    #Append the hardness and divide the energy by 2 to take into account the 2 in the formula of the LRF
    length_basis = len(basis)



    if nprocs >1:
        overlap_matrix = calculate_overlap_matrix_multithread(self,basis,nprocs)
    else:
        overlap_matrix = np.zeros((length_basis,length_basis))
        for first_transition in range(length_basis):
            for second_transition in range(first_transition+1):
                overlap_integral = self.mol_grids.integrate_product(transition_density_list[:,first_transition],overlap_matrix[:,second_transition])
                overlap_matrix[first_transition,second_transition] = overlap_integral
                overlap_matrix[second_transition,first_transition] = overlap_integral


    # s_matrix = np.einsum("k,ik,kj->ij",1/factor,overlap_matrix,overlap_matrix)


    # eigvalB,eigvecB = np.linalg.eigh(overlap_matrix)
    # eigvecBtilde = eigvecB * eigvalB**(-1/2)
    # Atilde = eigvecBtilde.transpose() @ s_matrix @ eigvecBtilde
    # eigenvalues, eigvecA = np.linalg.eigh(Atilde)
    # eigenvectors = eigvecBtilde @ eigvecA
    # overlap_matrix_inv = np.linalg.inv(overlap_matrix)
    # diag_matrix = overlap_matrix_inv.dot(s_matrix)
    #
    # eigenvalues, eigenvectors = np.linalg.eigh(diag_matrix)

    # eigenvalues, eigenvectors = sp.linalg.eigh(s_matrix,overlap_matrix)
    D = np.diag(1/factor).dot(overlap_matrix)
    eigenvalues,eigenvectors = np.linalg.eig(D)
    eigenvectors = eigenvectors.transpose()


    reconstructed_eigenvectors = []

    if nprocs >1:
        reconstructed_eigenvectors, reconstructed_eigenvector_norms = multithreading_reconstruct_eigenvectors_atomic_grids(self,basis,eigenvectors,nprocs)
    else:
        reconstructed_eigenvectors = []
        reconstructed_eigenvector_norms = []
        for eigenvector in eigenvectors:
            eigenvector_reshaped = eigenvector.reshape((N_trans,1,1,1))
            reconstructed_eigenvector = np.einsum("ijkl,ijkl->jkl",eigenvector_reshaped,basis)
            reconstructed_eigenvector_norm = mol.mol_grids.integrate_product(reconstructed_eigenvector,reconstructed_eigenvector)
            reconstructed_eigenvector_norms.append(reconstructed_eigenvector_norm)
            reconstructed_eigenvectors.append(reconstructed_eigenvector/reconstructed_eigenvector_norm**(1/2))

    eigenvectors = np.einsum("ij,i->ij",eigenvectors,1/np.array(reconstructed_eigenvector_norms)**(1/2))

    eig_sort = np.argsort(eigenvalues)
    eigenvalues = eigenvalues[eig_sort]
    eigenvectors = eigenvectors[eig_sort]
    reconstructed_eigenvectors = reconstructed_eigenvectors[eig_sort]
    self.properties["softness_kernel_eigenvectors"] = reconstructed_eigenvectors[::-1]
    self.properties["softness_kernel_eigenvalues"] = eigenvalues[::-1]
    self.properties["contribution_softness_kernel_eigenvectors"] = eigenvectors[::-1]


    time_taken = time.time() - time_before_calc

    print("######################################################")
    print("# Finished calculating eigenmodes of Softness Kernel #")
    print("# in {:3.3f} s #".format(time_taken))
    print("######################################################")

    return eigenvalues, reconstructed_eigenvectors



molecule.calculate_fukui_function = calculate_fukui_function
molecule.calculate_hardness = calculate_hardness
molecule.calculate_eigenmodes_linear_response_function = calculate_eigenmodes_linear_response_function
molecule.calculate_softness_kernel_eigenmodes = calculate_softness_kernel_eigenmodes


#####################################
## Functions to analyze eigenmodes ##
#####################################


def analysis_eigenmodes(self,kernel="linear_response_function",list_vectors=[1,2,3,4,5]):
    """
    analysis_eigenmodes(self,kernel="linear_response_function",list_vectors=[1,2,3,4,5])

    Computes some analysis of the eigenmodes of the linear_response_function or the softness_kernel.
    The kernels must be computed beforehand.
    The analysis are : Percentage of occupied MO in the mode, percentage of virtual MO in the mode.
    For the softness kernel, the proportion of electron transferred in each mode.
    These quantities will be stored in mol.properties["from_occ"], "to_virt", and "fukui_decomposition"
    :doc:`plot_analysis<../plot/plot_analysis>` can be used to plot these quantities

    Parameters
    ----------
        kernel : str
            The kernel on which the analysis will be made. Either "linear_reponse_function" or "softness_kernel"
        list_vectors : list, optional
            The list of the vectors that will be decomposed on Occupied Virtual orbitals. Starting to 1 instead of 0. By default [1,2,3,4,5]

    Returns
    -------
        None

    """

    if kernel == "linear_response_function":
        eigenvectors = self.properties["linear_response_eigenvectors"]
        eigenvalues = self.properties["linear_response_eigenvalues"]
        contrib_eigenvectors = self.properties["contribution_linear_response_eigenvectors"]
        transition_list = self.properties["transition_list"]
        transition_factor_list = self.properties["transition_factor_list"]

    elif kernel == "softness_kernel":
        eigenvectors = self.properties["softness_kernel_eigenvectors"]
        eigenvalues = self.properties["softness_kernel_eigenvalues"]
        contrib_eigenvectors = self.properties["contribution_softness_kernel_eigenvectors"]
        transition_list = self.properties["transition_list"] + [[[-1,-1]]]
        transition_factor_list = self.properties["transition_factor_list"] + [[[1]]]

    else:
        raise ValueError("The kernel {} is not known or is not implemented. The kernels available are linear_response_function and softness_kernel".format(kernel))

    lin_transition = np.array([transition for transitions in transition_list for transi in transitions for transition in transi])

    occ = lin_transition[::2]
    virt = lin_transition[1::2]
    max_occupied = np.max(occ)
    max_virtual= np.max(virt)


    from_occ_list = []
    to_virt_list = []
    for vector_number in list_vectors:
        if kernel=="softness_kernel":
            #To add fukui function
            transition_contrib = np.zeros((max_occupied+2,max_virtual+2))
            from_occ = np.zeros((max_occupied+2))
        else:
            transition_contrib = np.zeros((max_occupied+1,max_virtual+1))
            from_occ = np.zeros((max_occupied+1))

        contributions = contrib_eigenvectors[vector_number-1]

        for contrib in range(len(contributions)):
            transitions = transition_list[contrib]
            transition_factors = transition_factor_list[contrib]
            for transi,transi_factor in zip(transitions,transition_factors):
                transition_contrib[transi[0],transi[1]] += contributions[contrib] * transi_factor[0]

        for k in range(len(from_occ)):
            for j in range(len(transition_contrib[0])):
                from_occ[k] += transition_contrib[k,j]**2
        from_occ_list.append(from_occ)


        to_virt = np.zeros((max_virtual+1))
        for k in range(len(to_virt)):
            for j in range(len(transition_contrib)):
                to_virt[k] += transition_contrib[j,k]**2
        to_virt_list.append(to_virt)

    fukui_decomposition = []
    if kernel == "softness_kernel":
        for vector_number in range(len(eigenvalues)):
            contributions = contrib_eigenvectors[vector_number-1]
            fukui_decomposition.append(contributions[-1]**2 * self.properties["hardness"] * self.properties["softness_kernel_eigenvalues"][vector_number-1])
    f = np.array(fukui_decomposition)
    print(np.sum(f))



    self.properties["from_occ"] = from_occ_list
    self.properties["to_virt"] = to_virt_list
    if kernel == "softness_kernel":
        self.properties["fukui_decomposition"] = fukui_decomposition

    return


molecule.analysis_eigenmodes = analysis_eigenmodes


##TEMPO

#
# def calculate_overlaps(self):
#     """
#     calculate_overlaps()
#
#     Calculates the linear response function from the transition densities.
#     This method does not calculate directly the linear response, but only the eigenmodes.
#     The mathematics behind this function will soon be available somewhere...
#
#
#     Parameters
#     ----------
#     grid_points : list of 3
#     delta : float, optional
#         the length added on all directions of the box containing all atomic centers
#
#     Returns
#     -------
#     linear_response_eigenvalues
#         the eigenvalues of the linear response function
#     linear_response_eigenvectors
#         the eigenvectors of the linear response function
#
#     Notes
#     -----
#     The linear response function kernel can be computed as :
#
#     :math:`\\chi(r,r') = -2\\sum_{k\\neq0} \\frac{\\rho_0^k(r) \\rho_0^k(r')}{E_k-E_0}`
#
#     With :math:`\\rho_0^k` the transition density, and :math:`E_k` the energy of the transition k.
#
#     Therefore, the molecule needs the transition properties that can be extracted from a TD-DFT calculation, and the MO extracted from a molden file. More details can be found :doc:`here<../tuto/orbitals>`.
#     """
#
#     if not "transition_density_list" in self.properties:
#         self.calculate_transition_density()
#
#     #Dummy implementation for now
#     transition_density_list = self.properties["transition_density_list"]
#     transition_energy = np.array(self.properties['transition_energy'])
#
#     N_grids,N_trans,N_r,N_ang = np.shape(transition_density_list)
#
#     nprocs = rgmol.nprocs
#
#     print("#################################")
#     print("# Calculating Eigenmodes of LRF #")
#     print("#################################")
#     time_before_calc = time.time()
#
#     if nprocs >1:
#         transition_matrix = calculate_overlap_matrix_multithread(self,transition_density_list,nprocs)
#     else:
#         transition_matrix = np.zeros((N_trans,N_trans))
#         for first_transition in range(N_trans):
#             for second_transition in range(first_transition+1):
#                 overlap_integral = 0
#                 for grid,index_grid in zip(self.mol_grids.grids,range(N_grids)):
#                     overlap_integral += grid.integrate(transition_density_list[index_grid,first_transition] * transition_density_list[index_grid,second_transition])
#
#                 transition_matrix[first_transition,second_transition] = overlap_integral
#                 transition_matrix[second_transition,first_transition] = overlap_integral
#     D = np.diag(-2/transition_energy).dot(transition_matrix)
#
#     return D
#
# molecule.calculate_overlaps = calculate_overlaps